import pytest
import numpy as np
from unittest.mock import patch, MagicMock
from maidi import MidiScore
from maidi.integrations.api import MusicLangAPI
from maidi import instrument

API_URL = "https://api.mockmusiclang.io"
API_KEY = "testapikey"


@pytest.fixture
def setup_api():
    return MusicLangAPI(api_url=API_URL, api_key=API_KEY, verbose=True)


@pytest.fixture
def setup_score():
    instruments = [instrument.PIANO]
    score = MidiScore.from_empty(instruments=instruments, nb_bars=8, ts=(4, 4), tempo=120)
    return score


def test_determine_bars_step(setup_api):
    api = setup_api

    step = api._determine_bars_step(nb_added_bars_step=4, remaining_bars=10)
    assert step == 4

    step = api._determine_bars_step(nb_added_bars_step=None, remaining_bars=10)
    assert step == min(api.MAX_CONTEXT, 10)


def test_create_mask(setup_api, setup_score):
    api = setup_api
    score = setup_score
    score = score.add_silence_bars(8)
    mask = api._create_mask(score, 4)

    assert mask.shape == (score.nb_tracks, score.nb_bars)
    assert np.array_equal(mask[:, -4:], np.ones((score.nb_tracks, 4)))
    assert np.array_equal(mask[:, :-4], np.zeros((score.nb_tracks, score.nb_bars - 4)))


@patch.object(MusicLangAPI, 'predict')
def test_predict_segment(mock_predict, setup_api, setup_score):
    api = setup_api
    score = setup_score

    def predict_side_effect(predicted_score, mask, model, tags, chords, temperature, polling_interval, timeout,
                            prediction_kwargs):
        return predicted_score  # Return the input segment as the result

    mock_predict.side_effect = predict_side_effect

    predicted_score = score.add_silence_bars(api.MAX_CONTEXT)
    mask = api._create_mask(predicted_score, 4)
    subchords, subtags = None, None
    result = api._predict_segment(predicted_score, mask, model="control_masking_large", subtags=subtags,
                                  subchords=subchords,
                                  temperature=0.95, polling_interval=1, timeout=120, prediction_kwargs={})

    assert mock_predict.called
    assert result is not None
    assert result.nb_bars == predicted_score.nb_bars


@patch.object(MusicLangAPI, 'predict')
def test_extend(mock_predict, setup_api, setup_score):
    api = setup_api
    score = setup_score

    def predict_side_effect(predicted_score, mask, model, tags, chords, temperature, polling_interval, timeout,
                            prediction_kwargs):
        return predicted_score  # Return the input segment as the result

    mock_predict.side_effect = predict_side_effect

    extended_score = api.extend(score, nb_bars_added=4, model="control_masking_large", nb_added_bars_step=None,
                                chords=None, tags=None, timeout=120, temperature=0.95, polling_interval=1)

    assert mock_predict.called
    assert extended_score is not None
    assert extended_score.nb_bars == score.nb_bars + 4

    # Testing with nb_added_bars_step
    extended_score = api.extend(score, nb_bars_added=8, model="control_masking_large", nb_added_bars_step=4,
                                chords=None, tags=None, timeout=120, temperature=0.95, polling_interval=1)

    assert mock_predict.called
    assert extended_score is not None
    assert extended_score.nb_bars == score.nb_bars + 8


if __name__ == "__main__":
    pytest.main()



from maidi import MidiScore, INSTRUMENTS_DICT
from maidi import instrument

import pytest
import tempfile

def create_fake_midi_file(filename):
    import pretty_midi
    import numpy as np
    # Create a PrettyMIDI object
    midi_data = pretty_midi.PrettyMIDI()
    # Create an Instrument instance
    instrument1 = pretty_midi.Instrument(program=0)
    instrument2 = pretty_midi.Instrument(program=10)
    instruments = [instrument1, instrument2]

    pitches = [60, 64, 67]
    times = [(0, 1), (1, 2), (2, 3), (3, 4)]
    for instrument in instruments:
        # Iterate over note names, which will be converted to note number later
        for start, end in times:
            for pitch in pitches:
                # Create a Note instance, starting at 0s and ending at 1s
                note = pretty_midi.Note(
                    velocity=100, pitch=pitch, start=start, end=end)
                # Add it to our instrument
                instrument.notes.append(note)
        # Add the instrument to the PrettyMIDI object
        midi_data.instruments.append(instrument)
        # Write out the MIDI data
    midi_data.write(filename)

def test_empty_midi_score():
    """ """
    score = MidiScore.from_empty(
        instruments=[instrument.PIANO, instrument.VIOLIN],
        nb_bars=4,
        ts=(4, 4),
        tempo=120

    )

    assert score.shape == (2, 4)

    # Assert still same shape if writing and loading it
    new_score = MidiScore.from_base64(score.to_base64())

    assert new_score.shape == (2, 4)


def test_add_track():
    """ """
    score = MidiScore.from_empty(
        instruments=[instrument.PIANO, instrument.VIOLIN],
        nb_bars=4,
        ts=(4, 4),
        tempo=120

    )

    new_score = score.add_instrument(instrument.PIANO)

    assert new_score.shape == (3, 4)


def test_get_item_track_slice():
    """ """
    score = MidiScore.from_empty(
        instruments=[instrument.PIANO, instrument.VIOLIN],
        nb_bars=4,
        ts=(4, 4),
        tempo=120

    )
    new_score = score[1:]
    assert new_score.shape == (1, 4)

    program, is_drum = INSTRUMENTS_DICT[instrument.VIOLIN]
    assert new_score.track_keys == [(1, program, is_drum) ]
    assert len(new_score.tracks) == 1

    new_score = score[:1]
    assert new_score.shape == (1, 4)
    assert len(new_score.tracks) == 1
    program, is_drum = INSTRUMENTS_DICT[instrument.PIANO]
    assert new_score.track_keys == [(0, program, is_drum)]

    new_score = score[:]
    assert new_score.shape == (2, 4)
    assert len(new_score.tracks) == 2
    program1, is_drum1 = INSTRUMENTS_DICT[instrument.PIANO]
    program2, is_drum2 = INSTRUMENTS_DICT[instrument.VIOLIN]
    assert new_score.track_keys == [(0, program1, is_drum1), (1, program2, is_drum2)]

def test_get_item_chord_slice():
    """ """
    score = MidiScore.from_empty(
        instruments=[instrument.PIANO, instrument.VIOLIN],
        nb_bars=4,
        ts=(4, 4),
        tempo=120

    )
    new_score = score[:, 1:]
    assert new_score.shape == (2, 3)

    new_score = score[:, :1]
    assert new_score.shape == (2, 1)

    new_score = score[:, :]
    assert new_score.shape == (2, 4)


def test_set_item_single_track():
    score = MidiScore.from_empty(
        instruments=[instrument.PIANO, instrument.VIOLIN],
        nb_bars=4,
        ts=(4, 4),
        tempo=120
    )
    other_score = MidiScore.from_empty(
        instruments=[instrument.FLUTE],
        nb_bars=4,
        ts=(4, 4),
        tempo=120
    )

    # Modify the other track
    other_score.tracks[other_score.track_keys[0]][0]['time'][0] += 2

    score[0, :] = other_score
    assert score.track_keys[0] == (0, INSTRUMENTS_DICT[instrument.PIANO][0], INSTRUMENTS_DICT[instrument.PIANO][1] == 1)
    assert score.get_track_from_index(0) == other_score.get_track_from_index(0)


def test_set_item_multiple_slice():
    score = MidiScore.from_empty(
        instruments=[instrument.PIANO, instrument.VIOLIN],
        nb_bars=4,
        ts=(4, 4),
        tempo=120
    )
    other_score = MidiScore.from_empty(
        instruments=[instrument.FLUTE],
        nb_bars=3,
        ts=(4, 4),
        tempo=120
    )

    # Modify the other track
    other_score.tracks[other_score.track_keys[0]][2]['time'][0] += 2

    score[1:, 1:] = other_score
    assert score.track_keys[1] == (1, INSTRUMENTS_DICT[instrument.VIOLIN][0], INSTRUMENTS_DICT[instrument.VIOLIN][1] == 1)
    assert score[:, 1:].get_track_from_index(1) == other_score.get_track_from_index(0)

def test_set_item_invalid_type():
    score = MidiScore.from_empty(
        instruments=[instrument.PIANO, instrument.VIOLIN],
        nb_bars=4,
        ts=(4, 4),
        tempo=120
    )

    with pytest.raises(TypeError):
        score[0, 0] = "not a MidiScore"


from maidi import MidiScore
from maidi.analysis import TagsProvider


def get_tags_names():
    """ """
    return [
        "CONTROL_DENSITY__" + density
        for density in [
            "LOWEST",
            "LOWER",
            "LOW",
            "MEDIUM_LOW",
            "MEDIUM",
            "HIGH",
            "VERY_HIGH",
        ]
    ]


class DensityTagsProvider(TagsProvider):
    ALL_TAGS = get_tags_names()

    def get_tags(self, track_bar, chord, score):
        from maidi.utils.bar_helpers import bar_to_bar_duration_in_quarters
        notes = self.get_start_end_notes(track_bar, chord, score)
        bar_duration = bar_to_bar_duration_in_quarters(chord)
        return get_density_tags(notes, bar_duration)

def get_density_value(density):
    """From density in notes per quarters returns a density string in
    [LOWEST, LOWER, LOW, MEDIUM, MEDIUM_LOW HIGH, VERY_HIGH]
    :return:

    Parameters
    ----------
    density :
        

    Returns
    -------

    """
    if density < 0.34:
        return "LOWEST"
    if density < 0.5:
        return "LOWER"
    elif density < 2:
        return "LOW"
    elif density < 4:
        return "MEDIUM_LOW"
    elif density < 6:
        return "MEDIUM"
    elif density < 8:
        return "HIGH"
    else:
        return "VERY_HIGH"




def get_density(notes, bar_duration):
    """

    Parameters
    ----------
    notes :
        
    bar_duration :
        

    Returns
    -------

    """

    starts = [start for start, end in notes]
    if not starts:
        return "LOWEST"
    density = len(notes) / bar_duration
    return get_density_value(density)


def get_density_tags(notes, bar_duration):
    """

    Parameters
    ----------
    notes :
        
    bar_duration :
        

    Returns
    -------

    """
    density = get_density(notes, bar_duration)
    return ["CONTROL_DENSITY__" + density]





import numpy as np
from scipy.spatial.distance import cdist

from .chord_inference_utils import (
    TEMPLATES,
    COEFFS,
    CHROMA_VECTORS_MATRIX,
    PITCH_CLASSES_DICT,
    CHORD_TYPE_TO_PITCHES,
    CHORD_TYPES,
    PITCHES_DICT,
    PITCHES_DICT_ROOT,
)

EXTENSION_DICT = {
    (0, 3): "",
    (1, 3): "6",
    (2, 3): "64",
    (0, 4): "7",
    (1, 4): "65",
    (2, 4): "43",
    (3, 4): "2",
}


def max_correlation_index(
    input_matrix, template_matrix, bass_notes, use_first_max=False
):
    """

    Parameters
    ----------
    input_matrix :
        
    template_matrix :
        
    bass_notes :
        
    use_first_max :
         (Default value = False)

    Returns
    -------

    """

    # Add a small bonus to the bass note

    for i, bass_note in enumerate(bass_notes):
        if bass_note is not None:
            input_matrix[i, int(bass_note)] += 0.2

    # Calculate correlations
    correlations = 1 - cdist(input_matrix, template_matrix, "correlation")
    # Multiply the correlations by coefficients
    correlations *= COEFFS

    # Find index of max correlation for each row
    if use_first_max:
        max_corr_indices = np.argmax(correlations, axis=1)
        chord_type = max_corr_indices % len(TEMPLATES)
        chord_root = max_corr_indices // len(TEMPLATES)
        return chord_root, chord_type

    temp_len = len(TEMPLATES)
    max_corr_indicess = [
        (correlations[i] == np.max(correlations[i, :])).nonzero()[0].tolist()
        for i in range(len(correlations))
    ]
    chord_typess = [
        [i // temp_len for i in max_corr_indices]
        for max_corr_indices in max_corr_indicess
    ]
    chord_rootss = [
        [i % temp_len for i in max_corr_indices]
        for max_corr_indices in max_corr_indicess
    ]

    return chord_rootss, chord_typess


def get_chroma_vectors(pitches_bars):
    """

    Parameters
    ----------
    pitches_bars :
        

    Returns
    -------

    """
    # Remove drums
    pitches_class_array = [p % 12 for p in pitches_bars]
    value_counts = [
        np.bincount(pitches_class, minlength=12)
        for pitches_class in pitches_class_array
    ]
    # Get a 12xnb_bars matrix with each row being the indexes of the pitch class
    bar_chroma_vectors = np.asarray(value_counts)
    normalizer = np.sum(bar_chroma_vectors, axis=1)
    normalizer[normalizer == 0] = 1
    bar_chroma_vectors = bar_chroma_vectors / normalizer[:, None]
    return bar_chroma_vectors


def get_bass_note_bar(pitches_bars):
    """Return the bass note of each bar

    Parameters
    ----------
    notes :
        
    bars :
        
    pitches_bars :
        

    Returns
    -------

    
    """

    bass_notes = []
    for pitches in pitches_bars:
        if len(pitches) == 0:
            bass_notes.append(None)
        else:
            min_pitch = np.min(pitches) % 12
            bass_notes.append(min_pitch)
    return bass_notes


def dynamic_programming_for_chord_inference(lists):
    """

    Parameters
    ----------
    lists :
        

    Returns
    -------

    """

    n = len(lists)
    m = max(len(lst) for lst in lists)
    dp = [[float("inf")] * m for _ in range(n)]
    indices = [[0] * m for _ in range(n)]

    for i, s in enumerate(lists[0]):
        dp[0][i] = 0  # No string changes needed at the start

    for i in range(1, n):
        for j in range(len(lists[i])):
            # FIXME : Better cost function to include relative minor/major and key distances
            for k in range(len(lists[i - 1])):
                if (
                    lists[i][j] != lists[i - 1][k]
                ):  # If the strings are different, we need a change
                    cost = 1
                else:
                    cost = 0

                if (
                    dp[i][j] > dp[i - 1][k] + cost
                ):  # If we found a better (smaller) cost, update the dp and indices array
                    dp[i][j] = dp[i - 1][k] + cost
                    indices[i][j] = k

    res = [0] * n
    min_cost, min_index = min(
        (cost, index) for index, cost in enumerate(dp[-1])
    )  # Find the smallest cost in the last row
    res[-1] = min_index  # The last index is the one with the smallest cost

    for i in range(n - 2, -1, -1):  # Backtrack to find the indices for the other lists
        res[i] = indices[i + 1][res[i + 1]]

    return res


def get_pitch_class_set_tonality(tonality_root, tonality_mode):
    """

    Parameters
    ----------
    tonality_root :
        
    tonality_mode :
        

    Returns
    -------

    """
    major_scale = [0, 2, 4, 5, 7, 9, 11]
    minor_scale = [0, 2, 3, 5, 7, 8, 11]

    pitch_classes = major_scale if tonality_mode == "M" else minor_scale
    pitch_classes = [
        (pitch_class + tonality_root) % 12 for pitch_class in pitch_classes
    ]
    return set(pitch_classes)


def get_pitch_tonality_vector(tonality_root, tonality_mode):
    """

    Parameters
    ----------
    tonality_root :
        
    tonality_mode :
        

    Returns
    -------

    """
    tonality = get_pitch_class_set_tonality(tonality_root, tonality_mode)
    pitch_vector = [1.0 * (i in tonality) for i in range(12)]
    return pitch_vector


def optimal_chord_inference(candidates, chord_durations, bass_notes=None, bars=None):
    """List of candidates per bar

    Parameters
    ----------
    candidates :
        
    chord_durations :
        
    bass_notes :
         (Default value = None)
    bars :
         (Default value = None)

    Returns
    -------

    
    """
    candidates_nb = [
        [(c[1] + 3 * (c[2] == "m")) % 12 for c in candidate] for candidate in candidates
    ]

    indexes = dynamic_programming_for_chord_inference(candidates_nb)
    chords_props = [candidates[i][indexes[i]] for i in range(len(indexes))]

    if bars is None:
        bars = [None] * len(chords_props)
    if bass_notes is None:
        bass_notes = [None] * len(chords_props)

    chords = []
    for (
        (roman, tonality_root, tonality_mode, extension),
        chord_duration,
        bass_note,
    ) in zip(chords_props, chord_durations, bass_notes):
        num, den = chord_duration
        chord = (roman, tonality_root, tonality_mode, num, den, extension)

        if bass_note is not None:
            chord = get_chord_extended_from_bass_note(bass_note, chord)
        else:
            chord = (roman, tonality_root, tonality_mode, 0, extension, num, den)
        chords.append(chord)

    # Find chord octave
    return chords


def get_chord_pitch_classes(roman, tonality_root, tonality_mode, num, den, extension):
    """

    Parameters
    ----------
    roman :
        
    tonality_root :
        
    tonality_mode :
        
    num :
        
    den :
        
    extension :
        

    Returns
    -------

    """
    return PITCHES_DICT[(roman, tonality_root, tonality_mode, extension)]


def get_chord_octave(pitch):
    """

    Parameters
    ----------
    pitch :
        

    Returns
    -------

    """
    return -((pitch + 5) // 12)


def get_base_extension_from_inversion(inversion):
    """Get '' or '7' depending of the inversion (3 or 4 notes chords) and bass idx

    Parameters
    ----------
    inversion :
        return:

    Returns
    -------

    """
    three_notes = ["", "6", "64"]
    four_notes = ["7", "65", "43", "2"]
    is_three_notes = inversion in three_notes
    good_candidate_list = three_notes if is_three_notes else four_notes
    inversion_index = good_candidate_list.index(inversion)
    base_extension = "" if is_three_notes else "7"
    return base_extension, inversion_index


def get_octave_from_chord(roman, tonality_root, tonality_mode, extension):
    """Get the octave of the chord to minimize bass distance to central C

    Parameters
    ----------
    roman :
        param tonality_root:
    tonality_mode :
        param extension:
    tonality_root :
        
    extension :
        

    Returns
    -------

    """
    import re

    pattern = r"\([^)]*\)"
    anti_pattern = r"\(.*?\)"
    inversion = re.sub(pattern, "", extension)
    raw_extension = "".join(re.findall(anti_pattern, extension))
    base_extension, inversion_index = get_base_extension_from_inversion(inversion)
    real_extension = base_extension + raw_extension
    chord_pitches_real = PITCHES_DICT_ROOT[
        (roman, tonality_root, tonality_mode, real_extension)
    ]
    chord_octave = get_chord_octave(chord_pitches_real[inversion_index])
    return chord_octave


def get_chord_extended_from_bass_note(bass_note, chord):
    """

    Parameters
    ----------
    bass_note :
        
    chord :
        

    Returns
    -------

    """
    roman, tonality_root, tonality_mode, num, den, extension = chord
    chord_pitches_class = get_chord_pitch_classes(*chord)
    nb_notes = len(chord_pitches_class)

    chord_pitches_real = PITCHES_DICT_ROOT[
        (roman, tonality_root, tonality_mode, extension)
    ]
    if bass_note in chord_pitches_class:
        bass_idx = chord_pitches_class.index(bass_note)
        root_pitch = chord_pitches_real[bass_idx]
        new_extension = EXTENSION_DICT[(bass_idx, nb_notes)]
        real_extension = new_extension + (
            extension[1:] if extension.startswith("7") else extension
        )
    else:
        root_pitch = chord_pitches_real[0]
        real_extension = extension

    chord_octave = get_chord_octave(root_pitch)
    return roman, tonality_root, tonality_mode, chord_octave, real_extension, num, den


def get_nb_bars(tracks):
    """

    Parameters
    ----------
    tracks :
        

    Returns
    -------

    """
    return len(tracks[list(tracks.keys())[0]])


def concatenate_pitches(T):
    """

    Parameters
    ----------
    T :
        

    Returns
    -------

    """
    return np.concatenate(T) if len(T) > 0 else np.array([], dtype=int)


def fast_chord_inference(tracks, chord_durations):
    """Return the chord progression of the song, one chord per bar
    It uses dynamic programming to find the roman numeral chord progression that minimize the number of tonality changes

    Parameters
    ----------

    tracks : dict
        Dictionary of tracks
        
    chord_durations : list
        List of tuples (num, den) with the duration of each chord in quarters

    Returns
    -------
    list
        List of tuples (roman, tonality_root, tonality_mode, chord_octave, extension, time signature num, time signature den) with the chord progression

    """

    nb_bars = get_nb_bars(tracks)
    pitches_bars = [
        concatenate_pitches(
            [
                track_bar[i]["pitch"]
                for (idx, program, is_drum), track_bar in tracks.items()
                if not is_drum
            ]
        )
        for i in range(nb_bars)
    ]
    bar_chroma_vectors = get_chroma_vectors(pitches_bars)
    # Get the lowest note of each bar
    bass_notes = get_bass_note_bar(pitches_bars)
    chord_rootss, chord_typess = max_correlation_index(
        bar_chroma_vectors, CHROMA_VECTORS_MATRIX, bass_notes, use_first_max=False
    )

    candidates = []
    for chord_roots, chord_types, chroma in zip(
        chord_rootss, chord_typess, bar_chroma_vectors
    ):
        all_sub_candidates = []
        for chord_type, chord_root in zip(chord_roots, chord_types):
            # Get the list of roman numeral candidates
            pitches = tuple(
                sorted(
                    [
                        (chord_root + pitch) % 12
                        for pitch in CHORD_TYPE_TO_PITCHES[CHORD_TYPES[chord_type]]
                    ]
                )
            )
            subcandidates = PITCH_CLASSES_DICT[pitches]
            # Filter only maximum correlation candidates
            pitch_vectors = np.asarray(
                [
                    get_pitch_tonality_vector(tonality_root, tonality_mode)
                    for (
                        roman,
                        tonality_root,
                        tonality_mode,
                        extension,
                    ) in subcandidates
                ]
            )
            correlations = np.asarray(
                [np.correlate(chroma, pitch_vector) for pitch_vector in pitch_vectors]
            )
            # Get the list of maximum correlation candidates (there can be several)
            max_correlations = np.where(correlations == correlations.max())[0]
            subcandidates = [subcandidates[i] for i in max_correlations]
            all_sub_candidates += subcandidates

        if len(all_sub_candidates) == 0:
            all_sub_candidates = [(0, 0, "M", "")]
        candidates.append(all_sub_candidates)

    # Choose the candidate chord path that minimize the number of tonality changes
    chords = optimal_chord_inference(candidates, chord_durations, bass_notes=bass_notes)
    return chords



MODEL_CONTROL_MASKING_LARGE = "control_masking_large"
MODEL_MASKING_LARGE = "masking_large"
MODEL_MASKING_SMALL = "masking_small"

MODELS = [MODEL_CONTROL_MASKING_LARGE, MODEL_MASKING_LARGE, MODEL_MASKING_SMALL]



import pytest
from unittest.mock import patch
from maidi.analysis.control_features.special_notes import _get_special_notes_tags

# Define some constants for test cases
TRACK_BAR_1 = {
    "pitch": [60, 62, 64, 65],
    "velocity": [90, 85, 80, 75],
    "time": [0, 1, 2, 3],
    "duration": [1, 1, 1, 1],
}

TRACK_BAR_2 = {
    "pitch": [60, 61, 62, 63],
    "velocity": [90, 85, 80, 75],
    "time": [0, 1, 2, 3],
    "duration": [2, 2, 1, 1],
}

TRACK_BAR_3 = {
    "pitch": [60, 61, 62, 63, 64, 65],
    "velocity": [90, 85, 80, 75, 70, 65],
    "time": [0, 1, 2, 3, 4, 5],
    "duration": [1, 1, 1, 1, 1, 1],
}

CHORD = "Cmaj7"


@pytest.fixture
def mock_parse_pitch():
    """ """
    with patch("maidi.analysis.control_features.special_notes.parse_pitch") as mock:
        yield mock


def test_get_special_notes_tags_case_1(mock_parse_pitch):
    """

    Parameters
    ----------
    mock_parse_pitch :
        

    Returns
    -------

    """
    # Mocking parse_pitch to return some interesting and some uninteresting values
    mock_parse_pitch.side_effect = [
        ("s", 1, 4),  # Ninth in the scale (interesting)
        ("s", 3, 4),  # Fourth (interesting)
        ("s", 0, 4),  # Not interesting
        ("s", 0, 4),  # Not interesting
    ]

    result = _get_special_notes_tags(TRACK_BAR_1, CHORD)
    assert result == [("s", 1), ("s", 3)]


def test_get_special_notes_tags_case_2(mock_parse_pitch):
    """

    Parameters
    ----------
    mock_parse_pitch :
        

    Returns
    -------

    """
    mock_parse_pitch.side_effect = [
        ("s", 1, 4),  # Ninth in the scale (interesting)
        ("h", 1, 4),  # b9 (interesting)
        ("h", 2, 4),  # 9 (chromatic) (interesting)
        ("s", 0, 4),  # Not interesting
    ]

    result = _get_special_notes_tags(TRACK_BAR_2, CHORD)
    assert result == [("s", 1), ("h", 1), ("h", 2)]


def test_get_special_notes_tags_case_3(mock_parse_pitch):
    """

    Parameters
    ----------
    mock_parse_pitch :
        

    Returns
    -------

    """
    mock_parse_pitch.side_effect = [
        ("s", 1, 4),  # Ninth in the scale (interesting)
        ("s", 3, 4),  # Fourth (interesting)
        ("s", 0, 4),  # Not interesting
        ("s", 0, 4),  # Not interesting
        ("s", 5, 4),  # Sixth (interesting)
        ("s", 0, 4),  # Not interesting
    ]

    result = _get_special_notes_tags(TRACK_BAR_3, CHORD)
    assert result == [
        ("s", 1),
        ("s", 3),
    ]  # Top 3 should be returned, but s0 in top 3 so not returned


def test_get_special_notes_tags_case_with_normalization(mock_parse_pitch):
    """

    Parameters
    ----------
    mock_parse_pitch :
        

    Returns
    -------

    """
    mock_parse_pitch.side_effect = [
        ("s", 1, 4),  # Ninth in the scale (interesting)
        ("s", 0, 4),  # Not interesting
        ("h", 2, 4),  # 9 (chromatic) (interesting)
        ("s", 1, 4),  # Ninth in the scale (interesting)
    ]

    track_bar = {
        "pitch": [60, 61, 62, 60],
        "velocity": [90, 85, 80, 75],
        "time": [0, 1, 2, 3],
        "duration": [2, 2, 2, 4],  # The last note has the longest duration
    }

    result = _get_special_notes_tags(track_bar, CHORD)
    assert result == [("s", 1), ("h", 2)]


def test_get_special_notes_tags_case_empty_bar(mock_parse_pitch):
    """

    Parameters
    ----------
    mock_parse_pitch :
        

    Returns
    -------

    """
    mock_parse_pitch.side_effect = []

    track_bar = {"pitch": [], "velocity": [], "time": [], "duration": []}

    result = _get_special_notes_tags(track_bar, CHORD)
    assert result == []


if __name__ == "__main__":
    pytest.main()



import pytest
import numpy as np
from unittest.mock import patch
from maidi import MidiScore
from maidi.integrations.api import MusicLangAPI

API_URL = "https://api.mockmusiclang.io"
API_KEY = "testapikey"

@pytest.fixture
def setup_api():
    return MusicLangAPI(api_url=API_URL, api_key=API_KEY, verbose=True)

@pytest.fixture
def setup_score():
    instruments = ['piano']
    return MidiScore.from_empty(instruments=instruments, nb_bars=4, ts=(4, 4), tempo=120)

def test_initialization(setup_api):
    api = setup_api
    assert api.api_url == API_URL
    assert api.api_key == API_KEY
    assert api.verbose is True

def test_invalid_model(setup_api, setup_score):
    api = setup_api
    score = setup_score
    mask = np.ones((1, 4))
    with pytest.raises(ValueError) as excinfo:
        api.predict(score, mask, model="invalid_model")
    assert "Model invalid_model not existing" in str(excinfo.value)

def test_invalid_temperature(setup_api, setup_score):
    api = setup_api
    score = setup_score
    mask = np.ones((1, 4))
    with pytest.raises(ValueError) as excinfo:
        api.predict(score, mask, temperature=1.5)
    assert "Temperature must be lower than 1.0" in str(excinfo.value)

def test_invalid_mask_dimension(setup_api, setup_score):
    api = setup_api
    score = setup_score
    mask = np.ones((2, 4))  # Invalid mask shape
    with pytest.raises(ValueError) as excinfo:
        api.predict(score, mask)
    assert "Wrong number of tracks in the mask" in str(excinfo.value)

    mask = np.ones((1, 5))  # Invalid mask shape
    with pytest.raises(ValueError) as excinfo:
        api.predict(score, mask)
    assert "Wrong number of bars in the mask" in str(excinfo.value)

    mask = np.zeros((1, 4))  # Mask should not have only 0
    with pytest.raises(ValueError) as excinfo:
        api.predict(score, mask)
    assert "Mask must have at least one 1" in str(excinfo.value)

def test_invalid_chords(setup_api, setup_score):
    api = setup_api
    score = setup_score
    mask = np.ones((1, 4))
    invalid_chords = [(8, 0, 'M', '')]  # Invalid chord degree
    with pytest.raises(ValueError) as excinfo:
        api.predict(score, mask, chords=invalid_chords)
    assert "Chord degree is wrong" in str(excinfo.value)

@patch.object(MusicLangAPI, '_predict_with_api')
def test_predict_call(mock_predict_with_api, setup_api, setup_score):
    api = setup_api
    score = setup_score
    mask = np.ones((1, 4))
    api.predict(score, mask, model="control_masking_large")
    assert mock_predict_with_api.called

@patch.object(MusicLangAPI, '_predict_with_api')
def test_predict_with_tags(mock_predict_with_api, setup_api, setup_score):
    api = setup_api
    score = setup_score
    mask = np.ones((1, 4))
    tags = [[['CONTROL_MIN_POLYPHONY__1'] for _ in range(4)] for _ in range(1)]
    api.predict(score, mask, tags=tags)
    assert mock_predict_with_api.called

@patch.object(MusicLangAPI, '_predict_with_api')
def test_predict_with_chords(mock_predict_with_api, setup_api, setup_score):
    api = setup_api
    score = setup_score
    mask = np.ones((1, 4))
    chords = [(0, 0, 'M', '') for _ in range(4)]
    api.predict(score, mask, chords=chords)
    assert mock_predict_with_api.called



from maidi.analysis import ScoreTagger
from maidi.analysis import tags_providers
from maidi import MidiScore

import tempfile

def create_fake_midi_file(filename):
    import pretty_midi
    # Create a PrettyMIDI object
    midi_data = pretty_midi.PrettyMIDI()
    # Create an Instrument instance
    instrument1 = pretty_midi.Instrument(program=0)
    instrument2 = pretty_midi.Instrument(program=10)
    instruments = [instrument1, instrument2]

    pitches = [60, 64, 67]
    times = [(0, 1), (1, 2), (2, 3), (3, 4)]
    for instrument in instruments:
        # Iterate over note names, which will be converted to note number later
        for start, end in times:
            for pitch in pitches:
                # Create a Note instance, starting at 0s and ending at 1s
                note = pretty_midi.Note(
                    velocity=100, pitch=pitch, start=start, end=end)
                # Add it to our instrument
                instrument.notes.append(note)
        # Add the instrument to the PrettyMIDI object
        midi_data.instruments.append(instrument)
        # Write out the MIDI data
    midi_data.write(filename)


def test_basic_density_tags_with_score_tagger():
    with tempfile.NamedTemporaryFile(suffix='.mid', delete=True) as f:
        create_fake_midi_file(f.name)
        score_tagger = ScoreTagger([
            tags_providers.DensityTagsProvider()
        ])

        score = MidiScore.from_midi(f.name)
        tags = score_tagger.tag_score(score)

        assert len(tags) == 2
        assert len(tags[0]) == 2
        assert len(tags[0][0]) == 1


def test_empty_score_lead_to_0_polyphony():  # Check that tagger remove low velocity notes
    score = MidiScore.from_empty(['piano', 'flute'], nb_bars=4, ts=(4, 4), tempo=120)
    tagger = ScoreTagger(
        [
            tags_providers.MinMaxPolyphonyTagsProvider(),
        ]
    )

    tags = tagger.tag_score(score)
    assert 'CONTROL_MAX_POLYPHONY__0' in tags[0][0]
    assert 'CONTROL_MIN_POLYPHONY__0' in tags[0][0]


import pytest
from maidi.analysis.control_features import (
    get_density_tags,

)

from maidi.analysis.control_features.density import (
    get_density_value,
    get_density,
    get_tags_names, DensityTagsProvider,
)


def test_get_density_value():
    """ """
    assert get_density_value(0.2) == "LOWEST"
    assert get_density_value(0.4) == "LOWER"
    assert get_density_value(1.5) == "LOW"
    assert get_density_value(3.0) == "MEDIUM_LOW"
    assert get_density_value(5.0) == "MEDIUM"
    assert get_density_value(7.0) == "HIGH"
    assert get_density_value(8.5) == "VERY_HIGH"


def test_get_density():
    """ """
    # Test case with no notes
    assert get_density([], 4) == "LOWEST"

    # Test case with notes leading to 'LOWEST' density
    notes = [(0, 1), (2, 3), (4, 5)]
    assert get_density(notes, 16) == "LOWEST"  # density = 3/16 = 0.1875

    # Test case with notes leading to 'LOWER' density
    notes = [(0, 1), (2, 3), (4, 5)]
    assert get_density(notes, 6) == "LOW"  # density = 3/6 = 0.5

    # Test case with notes leading to 'LOW' density
    notes = [(0, 1), (0.5, 1.5), (1, 2)]
    assert get_density(notes, 2) == "LOW"  # density = 3/2 = 1.5

    # Test case with notes leading to 'MEDIUM_LOW' density
    notes = [(0, 1), (0.25, 1.25), (0.5, 1.5), (0.75, 1.75)]
    assert get_density(notes, 2) == "MEDIUM_LOW"  # density = 4/2 = 2

    # Test case with notes leading to 'MEDIUM' density
    notes = [(0, 1), (0.2, 1.2), (0.4, 1.4), (0.6, 1.6), (0.8, 1.8), (1, 2)]
    assert get_density(notes, 2) == "MEDIUM_LOW"  # density = 6/2 = 3

    # Test case with notes leading to 'HIGH' density
    notes = [
        (0, 0.5),
        (0.1, 0.6),
        (0.2, 0.7),
        (0.3, 0.8),
        (0.4, 0.9),
        (0.5, 1),
        (0.6, 1.1),
        (0.7, 1.2),
    ]
    assert get_density(notes, 2) == "MEDIUM"  # density = 8/2 = 4

    # Test case with notes leading to 'VERY_HIGH' density
    notes = [
        (0, 0.25),
        (0.1, 0.35),
        (0.2, 0.45),
        (0.3, 0.55),
        (0.4, 0.65),
        (0.5, 0.75),
        (0.6, 0.85),
        (0.7, 0.95),
        (0.8, 1.05),
        (0.9, 1.15),
    ]
    assert get_density(notes, 2) == "MEDIUM"  # density = 10/2 = 5


def test_get_token_names():
    """ """
    expected_token_names = [
        "CONTROL_DENSITY__LOWEST",
        "CONTROL_DENSITY__LOWER",
        "CONTROL_DENSITY__LOW",
        "CONTROL_DENSITY__MEDIUM_LOW",
        "CONTROL_DENSITY__MEDIUM",
        "CONTROL_DENSITY__HIGH",
        "CONTROL_DENSITY__VERY_HIGH",
    ]
    assert get_tags_names() == expected_token_names


def test_density_provider_tags_list():
    """ """
    provider = DensityTagsProvider()
    assert provider.get_tags_list() == get_tags_names()


if __name__ == "__main__":
    pytest.main()




from maidi.midi_library import get_list_of_midi_files,get_midi_file
from maidi import MidiScore, instrument
from maidi.constants import INSTRUMENTS_DICT


def test_horizontal_concatenation():
    midi_file = get_list_of_midi_files()[0]
    score1 = MidiScore.from_midi(get_midi_file(midi_file))
    score2 = MidiScore.from_midi(get_midi_file(midi_file))

    shape1 = score1.shape

    new_score = score1.concatenate(score2, axis=1)

    assert new_score.shape[1] == shape1[1] * 2

    assert new_score.bars == score1.bars + score2.bars

def test_vertical_concatenation():
    midi_file = get_list_of_midi_files()[0]
    score1 = MidiScore.from_midi(get_midi_file(midi_file))
    score2 = MidiScore.from_midi(get_midi_file(midi_file))

    shape1 = score1.shape

    new_score = score1.concatenate(score2, axis=0)

    assert new_score.shape[0] == shape1[0] * 2


def test_vertical_concatenation_good_instruments():
    score1 = MidiScore.from_empty(instruments=[instrument.VIOLIN, instrument.CELLO],
                                  nb_bars=4)
    score2 = MidiScore.from_empty(instruments=[instrument.PIANO, instrument.FLUTE],
                                  nb_bars=score1.nb_bars )

    shape1 = score1.shape
    nb_instruments = shape1[0]
    new_score = score1.concatenate(score2, axis=0)

    assert new_score.shape[0] == shape1[0] + score2.shape[0]
    assert new_score.track_keys[-2:] == [(nb_instruments, *INSTRUMENTS_DICT[instrument.PIANO]),
                                         (nb_instruments + 1, *INSTRUMENTS_DICT[instrument.FLUTE])]


def test_horizontal_concatenation_good_instruments():
    score1 = MidiScore.from_empty(instruments=[instrument.VIOLIN, instrument.CELLO],
                                  nb_bars=4)
    score2 = MidiScore.from_empty(instruments=[instrument.PIANO, instrument.FLUTE],
                                  nb_bars=4 )

    shape1 = score1.shape
    new_score = score1.concatenate(score2, axis=1)

    assert new_score.shape[1] == shape1[1] + score2.shape[1]
    assert new_score.track_keys == [(0, *INSTRUMENTS_DICT[instrument.VIOLIN]),
                                         (1, *INSTRUMENTS_DICT[instrument.CELLO])]




def bar_to_bar_duration_in_ticks(bar, tpq=24):
    """
    From a bar, returns the duration of a bar in ticks
    :param bar:
    :return:
    """
    from maidi import MidiScore

    ts_num = bar[MidiScore.TIME_SIGNATURE_NUMERATOR_INDEX]
    ts_den = bar[MidiScore.TIME_SIGNATURE_DENOMINATOR_INDEX]
    return tpq * 4 * ts_num / ts_den

def bar_to_bar_duration_in_quarters(bar):
    """
    From a bar, returns the duration of a bar in ticks
    :param bar:
    :return:
    """
    from maidi import MidiScore

    ts_num = bar[MidiScore.TIME_SIGNATURE_NUMERATOR_INDEX]
    ts_den = bar[MidiScore.TIME_SIGNATURE_DENOMINATOR_INDEX]
    return 4 * ts_num / ts_den


import pytest
import numpy as np
from unittest.mock import patch, MagicMock
from maidi import MidiScore
from maidi.integrations.api import MusicLangAPI
from maidi import instrument

API_URL = "https://api.mockmusiclang.io"
API_KEY = "testapikey"

@pytest.fixture
def setup_api():
    return MusicLangAPI(api_url=API_URL, api_key=API_KEY, verbose=True)

@pytest.fixture
def setup_scores():
    instruments = [instrument.PIANO]
    score1 = MidiScore.from_empty(instruments=instruments, nb_bars=8, ts=(4, 4), tempo=120)
    score2 = MidiScore.from_empty(instruments=instruments, nb_bars=8, ts=(4, 4), tempo=120)
    return score1, score2

def test_validate_transition_params(setup_api, setup_scores):
    api = setup_api
    score1, score2 = setup_scores
    score1 = score1.copy()
    score2 = score2.copy()
    api._validate_transition_params(score1, score2, 4)

    score2_too_much_instruments = score2.add_instrument(instrument.CLEAN_GUITAR)
    with pytest.raises(ValueError) as excinfo:
        api._validate_transition_params(score1, score2_too_much_instruments, 4)
    assert "The number of tracks in score1 and score2 must be the same" in str(excinfo.value)

    score1_new_instrument = score1.change_instrument(0, instrument.ACOUSTIC_GUITAR)
    with pytest.warns(UserWarning, match="Instruments are different in the two scores"):
        api._validate_transition_params(score1_new_instrument, score2, 4)

    with pytest.raises(ValueError) as excinfo:
        api._validate_transition_params(score1, score2, 13)
    assert "Transition size must be less than" in str(excinfo.value)

def test_calculate_bars_to_take(setup_api, setup_scores):
    api = setup_api
    score1, score2 = setup_scores
    nb_bars_score1, nb_bars_score2 = api._calculate_bars_to_take(score1, score2, 4)
    assert nb_bars_score1 == 6
    assert nb_bars_score2 == 6

def test_create_transition_segment(setup_api, setup_scores):
    api = setup_api
    score1, score2 = setup_scores
    score1 = score1.copy()
    score2 = score2.copy()
    nb_bars_score1 = 6
    nb_bars_transition = 4
    nb_bars_score2 = 6
    segment = api._create_transition_segment(score1, score2, nb_bars_score1, nb_bars_transition, nb_bars_score2)
    assert segment.nb_bars == nb_bars_score1 + nb_bars_transition + nb_bars_score2

def test_create_transition_mask(setup_api, setup_scores):
    api = setup_api
    score1, score2 = setup_scores
    score1 = score1.copy()
    score2 = score2.copy()
    nb_bars_score1 = 6
    nb_bars_transition = 4
    segment = api._create_transition_segment(score1, score2, nb_bars_score1, nb_bars_transition, 6)
    mask = api._create_transition_mask(score1.nb_tracks, segment.nb_bars, nb_bars_score1, nb_bars_transition)
    assert np.array_equal(mask[:, nb_bars_score1:nb_bars_score1 + nb_bars_transition], np.ones((score1.nb_tracks, nb_bars_transition)))
    assert np.array_equal(mask[:, :nb_bars_score1], np.zeros((score1.nb_tracks, nb_bars_score1)))
    assert np.array_equal(mask[:, nb_bars_score1 + nb_bars_transition:], np.zeros((score1.nb_tracks, 6)))

@patch.object(MusicLangAPI, 'predict')
def test_create_transition(mock_predict, setup_api, setup_scores):
    api = setup_api
    score1, score2 = setup_scores
    score1 = score1.copy()
    score2 = score2.copy()

    def predict_side_effect(segment, mask, model, timeout, temperature, async_mode, polling_interval, chords, tags, **prediction_kwargs):
        return segment  # Return the input segment as the result

    mock_predict.side_effect = predict_side_effect

    result = api.create_transition(score1, score2, 4, return_all=True)
    assert mock_predict.called
    assert result is not None

    result = api.create_transition(score1, score2, 4, return_all=False)
    assert mock_predict.called
    assert result is not None


from setuptools import setup, find_packages
from pathlib import Path

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()
requirements = (this_directory / "requirements.txt").read_text().splitlines()

setup(
    name='maidi',
    version='0.12.15',
    author="Florian GARDIN",
    author_email="fgardin.pro@gmail.com",
    description="A python package for symbolic AI music inference",
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires=requirements,
    packages=find_packages(include='*'),
    package_data={'maidi': ['examples/*.mid', 'metadata/*.pkl']},
    include_package_data=True,
)


import os
import glob

# Get absolute parent dir of this file

absolute_parent_dir = os.path.dirname(os.path.abspath(__file__))
midi_test_dir = os.path.join(absolute_parent_dir, "examples/*.mid")

# List all midi examples files
midi_files = glob.glob(midi_test_dir)

midi_files_dict = {}
for file in midi_files:
    midi_files_dict[os.path.basename(file).replace('.mid', '')] = file


def get_list_of_midi_files():
    return list(midi_files_dict.keys())

def get_midi_file(name):
    return midi_files_dict[name]


from symusic import Track, TimeSignature, Note, Score, Tempo
from copy import deepcopy
import numpy as np
import os
import tempfile
import requests
import base64
import time

from maidi.constants import INSTRUMENTS_DICT, API_URL_VARIABLE, API_KEY_VARIABLE


class MidiScore:
    """
    The main object to handle midi scores. It allows to manipulate midi scores in a numpy-like way by splitting it
    into two dimensions : tracks and bars.

    warning :: All midi notes with velocity < 2 are ignored when writing midi files, they are considered as fake notes. This behaviour will be removed in the future

    warning:: This class should not be directly instantiated, use :meth:`~MidiScore.from_midi()` or :meth:`~MidiScore.from_base64()` instead

    **Usage**

    Load a midi file and write it to a new file

    >>> from maidi import MidiScore
    >>> score = MidiScore.from_midi("path/to/midi.mid")
    >>> score.write("path/to/output.mid")

    Add a track and assign the content of the track to be the same as the first track

    >>> from maidi import MidiScore, instrument
    >>> score = MidiScore.from_midi("path/to/midi.mid")
    >>> score = score.add_instrument(instrument.ACOUSTIC_GUITAR)
    >>> score[-1, :] = score[0, :]

    Concatenate two scores horizontally

    >>> from maidi import MidiScore
    >>> score1 = MidiScore.from_midi("path/to/midi1.mid")
    >>> score2 = MidiScore.from_midi("path/to/midi2.mid")
    >>> score = score1.concatenate(score2, axis=1)
    >>> score.nb_bars == score1.nb_bars + score2.nb_bars
    True


    """

    SCALE_DEGREE_INDEX = 0
    TONALITY_INDEX = 1
    MODE_INDEX = 2
    CHORD_OCTAVE_INDEX = 3
    CHORD_EXTENSION_INDEX = 4
    TIME_SIGNATURE_NUMERATOR_INDEX = 5
    TIME_SIGNATURE_DENOMINATOR_INDEX = 6
    FAKE_NOTE_VELOCITY = 1  # Should be < 2 to be ignored by the model
    FAKE_NOTE_DURATION = 2
    MINIMUM_VELOCITY = 2
    FAKE_NOTE_PITCH = 60

    def __init__(self, bars, tracks, track_keys, tempo, tpq=24, **kwargs):
        """
        Create a new MidiScore object, you usually never directly call the constructor.
        See :meth:`~MidiScore.from_midi()` or :meth:`~MidiScore.from_base64()` to create a new MidiScore object


        Parameters
        ----------

        bars : list of list
            List of bars in the score
        tracks : dict
            Dictionary of tracks in the score
        track_keys : list
            List of track keys
        tempo : int
            Tempo of the score
        tpq : int
            Ticks per quarter note
        kwargs : dict
            Additional arguments

        """
        self.bars = bars
        self.tracks = tracks
        self.track_keys = track_keys
        self.tempo = tempo
        self.tpq = tpq
        self.kwargs = kwargs

    @classmethod
    def from_midi(cls, midi_file, tpq=24, bar_range=None):
        """

        Parameters
        ----------
        midi_file :
            
        tpq :
             (Default value = 24)
        bar_range :
             (Default value = None)

        Returns
        -------

        """
        from maidi.parser import Parser

        parser = Parser(tpq=tpq)
        bars, tracks, track_keys, tempo = parser.parse(midi_file)
        score = cls(bars, tracks, track_keys, tempo, tpq=tpq)
        if bar_range is not None:
            score = score.get_score_between(*bar_range)
        return score

    @property
    def shape(self):
        """
        Return the shape of the score (n_tracks, n_bars)
        """
        return len(self.track_keys), len(self.bars)
    @property
    def nb_tracks(self):
        """
        Number of tracks (or instruments) in the score
        """
        return len(self.track_keys)

    @classmethod
    def from_base64(cls, base64_string, tpq=24):
        """Create a score from a base64 string

        Parameters
        ----------
        base64_string :
            str, base64 string
        tpq :
            int, ticks per quarter note (Default value = 24)

        Returns
        -------

        """
        with tempfile.NamedTemporaryFile(suffix=".mid", delete=True) as midi_file:
            midi_file.write(base64.b64decode(base64_string))
            midi_file.flush()
            score = cls.from_midi(midi_file.name, tpq=tpq)
        return score


    def __getitem__(self, item):
        """
        Get the subscore corresponding to the items like the score was a matrix.
        Query in the numpy style
        The only difference is that all query will output an object of the same number of dimensions as the original object (A score)


        Parameters
        ----------

        item : int, slice or tuple
            The index or slice of the score to get
            If a tuple is provided, the first element is the track index and the second element is the bar index

        Returns
        -------
        score: MidiScore
            The subscore corresponding to the query

        """
        score = self
        if isinstance(item, tuple):
            first_dim, second_dim = item
            subtrack_score = score.get_tracks(first_dim)
            return subtrack_score.get_bars(second_dim)

        if isinstance(item, (slice, int)):
            return score.get_tracks(item)

    def concatenate(self, other_score, axis=0):
        """
        Concatenate two scores along a specific axis (0 = horizontally, 1 = vertically)

        For 0 axis concatenation, the number of tracks must be the same, the bars from left score are kept
        For 1 axis concatenation, the number of bars must be the same, the tracks from the left score are kept

        Parameters
        ----------

        other_score : MidiScore
            The score to concatenate with
        axis : int (Default value = 0)
            The axis along which to concatenate (0 = horizontally, 1 = vertically)

        """

        # Check shapes are compatible
        if axis == 0:
            if self.nb_tracks == 0:
                return other_score.copy()
            if other_score.nb_tracks == 0:
                return self.copy()
            if self.nb_bars != other_score.nb_bars:
                raise ValueError("The number of bars must be the same for axis 0 concatenation")
            new_score = self.copy()
            nb_tracks_to_add = len(other_score.track_keys)
            for track_key in other_score.track_keys:
                new_score = new_score.add_track(track_key[1], track_key[2])
            new_score[-nb_tracks_to_add:, :] = other_score.copy()
            return new_score

        if axis == 1:
            if self.nb_bars == 0:
                return other_score.copy()
            if other_score.nb_bars == 0:
                return self.copy()

            right_bars = deepcopy(other_score.bars)
            if self.nb_tracks != other_score.nb_tracks:
                raise ValueError("The number of tracks must be the same for axis 1 concatenation")
            new_score = self.copy().add_silence_bars(len(right_bars))
            new_score[:, -len(right_bars):] = other_score.copy()
            # bars should be set to second score
            new_score.bars = self.bars + right_bars
            return new_score

    def normalize_track_indexes(self):
        """
        Normalize the track indexes to be in the range of the number of tracks
        """
        score = self.copy()
        new_track_keys = []
        track_mapping = {}
        for real_track_idx, track_key in enumerate(score.track_keys):
            track_mapping[track_key] = (real_track_idx, *track_key[1:])
            new_track_keys.append((real_track_idx, *track_key[1:]))
        score.track_keys = new_track_keys
        new_tracks = {}
        for track_key, track in score.tracks.items():
            new_tracks[track_mapping[track_key]] = track
        score.tracks = new_tracks
        return score

    def __setitem__(self, item, value):
        """
        Set the subscore corresponding to the items like the score was a matrix.
        Set in the numpy style
        The only difference is that all query will output an object of the same number of dimensions as the original object (A score)

        Parameters
        ----------
        item : int, slice or tuple
            The index or slice of the score to get
            If a tuple is provided, the first element is the track index and the second element is the bar index
        value : MidiScore
            The subscore to set, should have the same shape as the query

        Returns
        -------
        None
        """
        if not isinstance(value, MidiScore):
            raise TypeError("Value must be a MidiScore instance")

        if isinstance(item, int):
            # Single integer index implies updating a specific track across all tracks
            track_indexes = [item]

            # Ensure the value has the same structure
            if len(value.shape[0]) != 1:
                raise ValueError("The provided MidiScore must have one track")
            if len(value.shape[1]) != len(self.bars):
                raise ValueError("The provided MidiScore must have the same number of bars")

            # Update the track notes
            self.assign_tracks(track_indexes, value)

        elif isinstance(item, slice):
            start, end = item.start, item.stop
            if start is None and end is None:
                return value
            start = start if start is not None else 0
            end = end if end is not None else len(self.track_keys)

            if start < 0:
                start = len(self.track_keys) + start
            if end < 0:
                end = len(self.track_keys) + end + 1

            track_indexes = range(start, end)
            self.assign_tracks(track_indexes, value)

        if isinstance(item, tuple):

            first_dim, second_dim = item
            if isinstance(first_dim, int):
                first_dim = slice(first_dim, first_dim + 1 if first_dim + 1 != 0 else None )
            if isinstance(second_dim, int):
                second_dim = slice(second_dim, second_dim + 1 if second_dim + 1 != 0 else None)

            start_first_dim = first_dim.start if first_dim.start is not None else 0
            end_first_dim = first_dim.stop if first_dim.stop is not None else len(self.track_keys)
            start_second_dim = second_dim.start if second_dim.start is not None else 0
            end_second_dim = second_dim.stop if second_dim.stop is not None else len(self.bars)

            if start_first_dim < 0:
                start_first_dim = len(self.track_keys) + start_first_dim
            if start_second_dim < 0:
                start_second_dim = len(self.bars) + start_second_dim
            if end_first_dim < 0:
                end_first_dim = len(self.track_keys) + end_first_dim + 1
            if end_second_dim < 0:
                end_second_dim = len(self.bars) + end_second_dim + 1

            receiver_shape = (end_first_dim - start_first_dim, end_second_dim - start_second_dim)
            value_shape = value.shape
            if receiver_shape != value_shape:
                raise ValueError(f"Shapes do not match: {receiver_shape} != {value_shape}")

            self.assign_tracks(range(start_first_dim, end_first_dim), value, bar_range=(start_second_dim, end_second_dim))

        else:
            raise TypeError("Index must be an int or a tuple of two ints")

    def __iter__(self):
        return iter(self.track_keys)

    def __len__(self):
        return len(self.track_keys)

    def assign_tracks(self, track_indexes, other_score, bar_range=None):
        """
        Assign some tracks of the other_score to the current score
        """
        # Update the track notes
        for idx, track_index in enumerate(track_indexes):
            value_track = other_score.track_keys[idx]
            self_track = self.track_keys[track_index]
            if bar_range is None:
                for bar_idx in range(len(self.bars)):
                    self.tracks[self_track][bar_idx] = other_score.tracks[value_track][bar_idx]
            else:
                for abs_idx, bar_idx in enumerate(range(*bar_range)):
                    self.tracks[self_track][bar_idx] = other_score.tracks[value_track][abs_idx]

    def get_track_from_index(self, index):
        return self.tracks[self.track_keys[index]]

    def get_track_bar(self, idx_track, idx_bar):
        return self.tracks[self.track_keys[idx_track]][idx_bar]

    def get_chords_prompt(self):
        """
        Return the list of chords in the score in the format [(chord_degree, tonality, mode, chord extension), ...]

        Chord degree : int, degree of the chord in the tonality but 0 indexed (0-6), tonic is 0, fifth is 4
        Tonality : int, tonality in which the chord is played (0-11), 0 is C, 1 is C#, 2 is D, ...
        Mode : str, mode of the tonality, "m" for minor, "M" for major
        Chord extension : str, extension of the chord as in roman numeral notation, in ['', ',6', '64', '7', '65', '43', '2']
        Optionally you can use (sus2) or (sus4) for suspended chords

        Returns
        -------
        chords: list of tuple
        As specified above

        """
        kept_indexes = [self.SCALE_DEGREE_INDEX, self.TONALITY_INDEX, self.MODE_INDEX, self.CHORD_EXTENSION_INDEX]
        return [[bar[index] for index in kept_indexes] for bar in self.bars]
    def get_bars(self, item):
        """
        Get the bars between the start and end index

        slice: int or Slice, slice object
        """
        if isinstance(item, int):
            item = slice(item, item + 1 if item + 1 != 0 else None)
        start, end = item.start, item.stop

        if start is None and end is None:
            return self

        return self.get_score_between(start, end)


    def add_note(self, pitch, time, duration, velocity , track_index, bar_index):
        """Add a note to the score

        Parameters
        ----------
        pitch :
            int, pitch of the note
        velocity :
            int, velocity of the note
        duration :
            int, duration of the note
        track_index :
            int, index of the track
        bar_index :
            int, index of the bar
        time :
            int, time of the note

        Returns
        -------

        """
        score = self
        track_key = score.track_keys[track_index]
        score.tracks[track_key][bar_index]["time"] = np.append(score.tracks[track_key][bar_index]["time"], time)
        score.tracks[track_key][bar_index]["pitch"] = np.append(score.tracks[track_key][bar_index]["pitch"], pitch)
        score.tracks[track_key][bar_index]["duration"] = np.append(score.tracks[track_key][bar_index]["duration"], duration)
        score.tracks[track_key][bar_index]["velocity"] = np.append(score.tracks[track_key][bar_index]["velocity"], velocity)
        return score

    def get_tracks(self, item):
        """
        Get the tracks between the start and end index

        slice: int or Slice, slice object
        """
        if isinstance(item, int):
            item = slice(item, item + 1 if item + 1 != 0 else None)
        start, end = item.start, item.stop

        if start is None and end is None:
            return self

        return self.get_track_subset(start, end)


    def to_base64(self):
        """Transform the score to a base64 string
        :return:

        Parameters
        ----------

        Returns
        -------

        """
        with tempfile.NamedTemporaryFile(suffix=".mid", delete=True) as midi_file:
            self.write(midi_file.name)
            with open(midi_file.name, "rb") as f:
                encoded_string = base64.b64encode(f.read()).decode()
        # Remove the temporary file
        return encoded_string

    def delete_bar(self, bar_index):
        """Delete a bar from the score

        Parameters
        ----------
        bar_index :
            int, index of the bar to delete

        Returns
        -------

        """
        score = self.copy()
        new_bars = deepcopy(score.bars)
        new_tracks = deepcopy(score.tracks)

        for track_key in score.track_keys:
            idx, program, is_drum = track_key
            new_tracks[(idx, program, is_drum)].pop(bar_index)

        return MidiScore(new_bars, new_tracks, score.track_keys, score.tempo, score.tpq).normalize_track_indexes()

    @classmethod
    def _get_fake_note_vector(cls):
        """ """
        return {
            "time": np.asarray([0], dtype=np.int32),
            "pitch": np.asarray([cls.FAKE_NOTE_PITCH], dtype=np.int8),
            "duration": np.asarray([cls.FAKE_NOTE_DURATION], dtype=np.int32),
            "velocity": np.asarray([cls.FAKE_NOTE_VELOCITY], dtype=np.int8),
        }


    @classmethod
    def _get_bar_duration_from_bar(cls, bar, tpq):
        return int(bar[cls.TIME_SIGNATURE_NUMERATOR_INDEX] * 4 * tpq / bar[cls.TIME_SIGNATURE_DENOMINATOR_INDEX])

    def get_bar_duration(self, bar_index):
        """
        Get the duration of a bar in ticks for a specific bar of the score

        Parameters
        ----------
        bar_index : int
            Index of the bar



        Returns
        -------
        int
            Duration of the bar in ticks

        """
        return self._get_bar_duration_from_bar(self.bars[bar_index], self.tpq)

    @classmethod
    def _get_bar_start_end_tick_from_bars(cls, bars, bar_index, tpq):
        """
        Calculate the start and end tick of a given bar on a score

        """
        time = 0
        for temp_bar_idx, bar in enumerate(bars[:bar_index]):
            bar_duration = cls._get_bar_duration_from_bar(bar, tpq)
            time += bar_duration
        start = time
        end = time + cls._get_bar_duration_from_bar(bars[bar_index], tpq)

        return start, end

    def get_bar_start_end_tick(self, bar_index):
        """
        Calculate the start and end tick of a given bar on the score

        Parameters
        ----------
        bar_index :  int
            Index of the bar


        Returns
        -------
        start: int
            Start tick of the bar
        end: int
            End tick of the bar


        """
        return self._get_bar_start_end_tick_from_bars(self.bars, bar_index, self.tpq)

    def notes_to_note_tick_list(self, notes, bars):
        """

        Parameters
        ----------
        notes :
            
        bars :
            

        Returns
        -------

        """
        note_tick_list = []
        for idx, (bar, note_list) in enumerate(zip(bars, notes)):
            bar_start, bar_end = self.get_bar_start_end_tick(idx)
            for note in note_list:

                note_tick_list.append(
                    (
                        note.start,
                        note.end,
                        note.pitch,
                        bar[self.TIME_SIGNATURE_NUMERATOR_INDEX],
                        bar[self.TIME_SIGNATURE_DENOMINATOR_INDEX],
                        bar_start,
                        bar_end,
                        idx,
                    )
                )
        return note_tick_list

    def write(self, filename):
        """

        Parameters
        ----------
        filename :
            

        Returns
        -------

        """
        score = Score(self.tpq)
        score.tempos.append(Tempo(0, self.tempo))
        current_ts = None
        for index_bar, bar in enumerate(self.bars):
            start_bar, end_bar = self.get_bar_start_end_tick(index_bar)
            ts = TimeSignature(
                start_bar,
                bar[self.TIME_SIGNATURE_NUMERATOR_INDEX],
                bar[self.TIME_SIGNATURE_DENOMINATOR_INDEX],
            )
            candidate_ts = (ts.numerator, ts.denominator)
            if current_ts != candidate_ts:
                score.time_signatures.append(ts)
                current_ts = (
                    (ts.numerator, ts.denominator) if current_ts is None else current_ts
                )

        for track_key in self.track_keys:
            idx, program, is_drum = track_key
            track = Track(program=program, is_drum=is_drum)
            track.notes = self.track_notes_to_score_notes(
                self.bars, self.tracks[(idx, program, is_drum)], self.tpq
            )
            score.tracks.append(track)

        score.dump_midi(filename)
        return score

    @classmethod
    def track_notes_to_score_notes(cls, bars, track_notes, tpq):
        """

        Parameters
        ----------
        bars :
            
        track_notes :
            

        Returns
        -------

        """
        notes = []
        has_note = False
        for idx_bar, bar_track_notes in enumerate(track_notes):
            start_bar_tick, end_bar_tick = cls._get_bar_start_end_tick_from_bars(bars, idx_bar, tpq=tpq)
            for idx in range(len(bar_track_notes["time"])):
                time = (
                    bar_track_notes["time"][idx]
                    + start_bar_tick
                )
                pitch = bar_track_notes["pitch"][idx]
                duration = bar_track_notes["duration"][idx]
                velocity = bar_track_notes["velocity"][idx]
                notes.append(Note(time, duration, pitch, velocity))
                has_note = True
        if not has_note:
            notes.append(
                Note(
                    0,
                    cls.FAKE_NOTE_DURATION,
                    cls.FAKE_NOTE_PITCH,
                    cls.FAKE_NOTE_VELOCITY,
                )
            )
        return notes

    @staticmethod
    def offset_notes(offset, tracks, track_keys):
        """

        Parameters
        ----------
        offset :
            
        tracks :
            
        track_keys :
            

        Returns
        -------

        """
        new_tracks = {}
        for track_key in track_keys:
            idx, program, is_drum = track_key
            new_tracks[(idx, program, is_drum)] = deepcopy(
                tracks[(idx, program, is_drum)]
            )
            for notes in new_tracks[(idx, program, is_drum)]:
                notes["time"] = (
                    np.asarray(notes["time"].copy(), dtype=np.int32) + offset
                )
                assert all(
                    [time >= 0 for time in notes["time"]]
                ), f"Negative time found in track {track_key}"

        return new_tracks

    def offset_score(self, offset):
        """

        Parameters
        ----------
        offset :
            

        Returns
        -------

        """
        new_bars = deepcopy(self.bars)

        return MidiScore(
            new_bars,
            deepcopy(self.tracks),
            deepcopy(self.track_keys),
            self.tempo,
            self.tpq,
        )

    def get_score_between(self, start_bar, end_bar):
        """

        Parameters
        ----------
        start_bar :
            
        end_bar :
            

        Returns
        -------

        """

        new_bars = deepcopy(self.bars[start_bar:end_bar])
        if start_bar is None:
            start_bar = 0


        new_tracks = {}
        for track_key in self.track_keys:
            idx, program, is_drum = track_key
            new_tracks[(idx, program, is_drum)] = self.tracks[
                (idx, program, is_drum)
            ][start_bar:end_bar]
            assert len(new_tracks[(idx, program, is_drum)]) == len(
                new_bars
            ), "Track and bar length mismatch"
        return MidiScore(new_bars, new_tracks, self.track_keys, self.tempo, self.tpq)

    def copy(self):
        """ """
        return MidiScore(
            deepcopy(self.bars),
            deepcopy(self.tracks),
            deepcopy(self.track_keys),
            self.tempo,
            self.tpq,
        )

    def get_mask(self):
        """Get the mask of the score (np.array of shape (n_tracks, n_bars))
        The mask allows to specify which bar of the score should be regenerated by a model (ones)
        The mask is the same size as the score and is empty by default (all zeros)

        Parameters
        ----------

        Returns
        -------
        mask: np.array of shape (n_tracks, n_bars)
            Mask of the score, empty by default (all zeros)

        """
        import numpy as np

        mask = np.zeros((len(self.track_keys), len(self.bars)))

        return mask

    def get_empty_controls(self, prevent_silence=True):
        """Get the mask, the tags and the bars of the current score.
        The mask is a np.array of shape (n_tracks, n_bars)
        The tags is a list of list of list of size (n_tracks, n_bars, <variable size>)
        The bars is a list of size (n_bars,)

        Parameters
        ----------
        prevent_silence :
            bool, if True add a tag to each bar to explicitly prevent silence during generation (Default value = False)

        Returns
        -------

        """
        mask = self.get_mask()
        tags = [[[] for _ in range(mask.shape[1])] for _ in range(mask.shape[0])]
        tags = self.prevent_silence_tags(tags) if prevent_silence else tags
        chords = [None for _ in range(mask.shape[1])]
        return mask, tags, chords

    @classmethod
    def prevent_silence_tags(self, tags):
        """Prevent silence tags from being added to the tags

        Parameters
        ----------
        tags :
            list of list of list of str, tags to prevent silence

        Returns
        -------

        """
        for i in range(len(tags)):
            for j in range(len(tags[i])):
                if not tags[i][j]:
                    tags[i][j].append("CONTROL_MIN_POLYPHONY__1")
        return tags

    @property
    def nb_bars(self):
        """ """
        return len(self.bars)

    def get_track_subset(self, start, end):
        """

        Parameters
        ----------
        start :
            
        end :
            

        Returns
        -------

        """
        score = self.copy()
        score.track_keys = score.track_keys[start:end]
        score.tracks = {
            track_key: score.tracks[track_key] for track_key in score.track_keys
        }
        return score

    def get_silenced_mask(self):
        """

        Parameters
        ----------

        Returns
        -------
        type
            :return:

        """
        mask = np.zeros((len(self.track_keys), len(self.bars)))
        for i, track_key in enumerate(self.track_keys):
            idx, program, is_drum = track_key
            for j, bar in enumerate(self.bars):
                mask[i, j] = (
                    self.tracks[(idx, program, is_drum)][j]["pitch"].shape[0]
                    == 0
                )
        return mask

    def get_silenced_bars(self):
        """ """
        silenced_mask = self.get_silenced_mask()
        silenced_bars = np.all(silenced_mask == 1, axis=0)
        return silenced_bars

    def cut_silenced_bars_at_beginning_and_end(self):
        """Cut silenced bars at the beginning and end of the score
        :return:

        Parameters
        ----------

        Returns
        -------

        """
        score = self.copy()
        silenced_bars = score.get_silenced_bars()
        # Cut silenced bars at the beginning
        bar_to_cut_start = 0
        while silenced_bars[bar_to_cut_start]:
            bar_to_cut_start += 1
            if bar_to_cut_start == len(silenced_bars):
                break

        # Cut silenced bars at the end
        bar_to_cut_end = 0
        while silenced_bars[-bar_to_cut_end - 1]:
            bar_to_cut_end += 1
            if bar_to_cut_end == len(silenced_bars):
                break
        score = score.get_score_between(
            bar_to_cut_start, len(score.bars) - bar_to_cut_end
        )
        return score

    def check_mask(self, mask):
        """Check that the mask is valid

        Parameters
        ----------
        mask :
            np.array of shape (n_tracks, n_bars)

        Returns
        -------

        """
        if mask.shape[1] != len(self.bars):
            raise ValueError(
                f"Mask length does not match score length {mask.shape[1]} != {len(self.bars)}"
            )
        if mask.shape[1] > 16:
            raise ValueError("Mask length must be less than 16")
        if mask.shape[0] != len(self.track_keys):
            raise ValueError(
                "Mask number of tracks does not match score number of tracks"
            )
        if np.sum(mask) == 0:
            raise ValueError("Mask must have at least one 1")

    def check_times(self):
        """ """
        for track_key in self.track_keys:
            idx, program, is_drum = track_key
            assert len(self.tracks[(idx, program, is_drum)]) == len(
                self.bars
            ), "Track and bar length mismatch"

    def sanitize_score(self, score_to_compare, mask):
        """Do checks after predicting a score

        Parameters
        ----------
        score :
            param score_to_compare:
        mask :
            return:
        score_to_compare : MidiScore


        Returns
        -------

        """
        score = self.copy()
        score.check_times()
        if len(score.track_keys) > len(score_to_compare.track_keys):
            # If the model added tracks, we need to remove them
            score = score.get_track_subset(0, len(score_to_compare.track_keys))

        if score.nb_bars < mask.shape[1]:
            pass
            # score = score.add_silence_bars(mask.shape[1] - score.nb_bars, add_fake_notes=True)
        if score.nb_bars > mask.shape[1]:
            score = score.get_score_between(0, mask.shape[1])

        return score

    def cut_notes_to_end_of_bars(self):
        """Cut durations of notes that go beyond the end of the bar
        :return:

        Parameters
        ----------

        Returns
        -------

        """
        self = self.copy()
        new_tracks = {}
        for track_key in self.track_keys:
            idx, program, is_drum = track_key
            new_tracks[(idx, program, is_drum)] = deepcopy(
                self.tracks[(idx, program, is_drum)]
            )
            for idx_bar, bar in enumerate(self.bars):
                bar_start, bar_end = self.get_bar_start_end_tick(idx_bar)
                bar_duration = bar_end - bar_start
                for idx_note, note in enumerate(
                    new_tracks[(idx, program, is_drum)][idx_bar]["time"]
                ):
                    if (
                        note
                        + new_tracks[(idx, program, is_drum)][idx_bar][
                            "duration"
                        ][idx_note]
                        > bar_duration
                    ):
                        new_tracks[(idx, program, is_drum)][idx_bar][
                            "duration"
                        ][idx_note] = (bar_duration - note)

        return MidiScore(
            deepcopy(self.bars),
            new_tracks,
            deepcopy(self.track_keys),
            self.tempo,
            self.tpq,
        )

    def add_instrument(self, instrument):
        try:
            program, is_drum = INSTRUMENTS_DICT[instrument]
        except KeyError:
            raise ValueError(f"Unknown instrument {instrument}")

        return self.add_track(program, is_drum)

    def add_track(self, program, is_drum):
        """Add an empty track (with fake notes) to the score

        Parameters
        ----------
        program :
            int, program number
        is_drum :
            bool, is drum track

        Returns
        -------

        """
        self = self.copy()
        max_track_index = max([track_key[0] for track_key in self.track_keys])
        new_track_key = (max_track_index+1, program, bool(is_drum))
        self.track_keys.append(new_track_key)
        self.tracks[new_track_key] = [
            self._get_fake_note_vector() for _ in range(len(self.bars))
        ]
        return self.normalize_track_indexes()

    @classmethod
    def match_track_keys(cls, score, other_score):
        """Match the track keys of two scores
        First find the midi with the most tracks
        Assert shortest score tracks match the beginning of the longest score tracks
        Then add fake tracks to the shortest score
        And returns both tracks in the same order

        Parameters
        ----------
        other_score :
            MidiScore, Score on which to match track keys
        score :
            

        Returns
        -------

        """
        score = score.copy()
        other_score = other_score.copy()

        if score.track_keys == other_score.track_keys:
            return score, other_score

        first_longer = len(score.track_keys) > len(other_score.track_keys)
        if first_longer:
            longer_score = score
            shorter_score = other_score
        else:
            longer_score = other_score
            shorter_score = score

        # Assert that the shorter score tracks match the beginning of the longer score tracks
        for i, track_key in enumerate(shorter_score.track_keys):
            if track_key != longer_score.track_keys[i]:
                raise ValueError(
                    "Score A and B have totally different tracks, cannot match them"
                )
        # Add fake tracks to the shorter score
        for i in range(len(shorter_score.track_keys), len(longer_score.track_keys)):
            shorter_score = shorter_score.add_track(
                longer_score.track_keys[i][1], longer_score.track_keys[i][2]
            )

        return score, other_score

    def change_instrument(self, track_idx, instrument):
        """Change the instrument of a track

        Parameters
        ----------
        track_idx :
            param instrument:
        instrument :
            

        Returns
        -------

        """
        score = self.copy()
        track_key = score.track_keys[track_idx]
        new_track_index = (
            track_key[0],
            INSTRUMENTS_DICT[instrument][0],
            bool(INSTRUMENTS_DICT[instrument][1]),
        )
        score.track_keys[track_idx] = new_track_index
        # Change in tracks
        score.tracks[new_track_index] = score.tracks.pop(track_key)
        return score

    def add(self, other_score, silence_bars=0, try_matching_tracks=True):
        """Concatenate the current score with another score horizontally

        Parameters
        ----------
        other_score :
            MidiScore
        silence_bars :
            int, number of silence bars to add between the two scores (Default value = 0)
        try_matching_tracks :
            bool, if True try to match the tracks of the two scores (Default value = True)

        Returns
        -------

        """
        score = self.copy()
        if silence_bars > 0:
            initial_length = len(score.bars)
            score = score.add_silence_bars(silence_bars, add_fake_notes=False)
            assert (
                len(score.bars) == initial_length + silence_bars
            ), "Silence bars not added correctly"
        other_score = other_score.copy()
        # Assert that the other score has the same tracks
        if other_score.track_keys != score.track_keys:
            if try_matching_tracks:
                score, other_score = self.match_track_keys(score, other_score)
            else:
                raise ValueError("Track keys do not match")
        # Offset the other score to start after the current score
        new_bars = score.bars + other_score.bars
        new_tracks = {}
        for track_key in score.track_keys:
            idx, program, is_drum = track_key
            new_tracks[(idx, program, is_drum)] = (
                score.tracks[(idx, program, is_drum)]
                + other_score.tracks[(idx, program, is_drum)]
            )
        return MidiScore(
            new_bars, new_tracks, score.track_keys, score.tempo, score.tpq
        )

    def add_silence_bars(self, n_bars, add_fake_notes=True):
        """Add n_bars of silence to the end of the score

        Parameters
        ----------
        n_bars :
            int, number of silence bars to add
        add_fake_notes :
            bool, if True add a fake note to each track (Default value = True)

        Returns
        -------

        """
        score = self.copy()
        # Assert that the other score has the same tracks
        # Offset the other score to start after the current score
        if n_bars <= 0:
            raise ValueError(
                "n_bars must be positive"
            )  # If we want to add silence bars

        new_bars = score.bars
        for i in range(n_bars):
            new_bars.append(deepcopy(score.bars[-1]))

        new_tracks = {}
        for track_key in score.track_keys:
            idx, program, is_drum = track_key
            new_tracks[(idx, program, is_drum)] = score.tracks[
                (idx, program, is_drum)
            ]
            for new_bar in new_bars[-n_bars:]:
                if add_fake_notes:
                    new_tracks[(idx, program, is_drum)].append(
                        self._get_fake_note_vector()
                    )
                else:
                    new_tracks[(idx, program, is_drum)].append(
                        {
                            "time": np.asarray([], dtype=np.int32),
                            "pitch": np.asarray([], dtype=np.int8),
                            "duration": np.asarray([], dtype=np.int32),
                            "velocity": np.asarray([], dtype=np.int8),
                        }
                    )
            # Assert good length
            assert len(new_tracks[(idx, program, is_drum)]) == len(
                new_bars
            ), "Track and bar length mismatch"
        return MidiScore(
            new_bars, new_tracks, score.track_keys, score.tempo, score.tpq
        )

    @property
    def instruments(self):
        """
        Get the list of instruments in the score as a list of strings

        Returns
        -------
        instruments: list of str

        """
        from maidi.constants import REVERSE_INSTRUMENT_DICT
        return [REVERSE_INSTRUMENT_DICT[(track_key[1], track_key[2])] for track_key in self.track_keys]

    @classmethod
    def from_empty(cls, instruments, nb_bars, ts=(4, 4), tempo=120, tpq=24):
        """

        Parameters
        ----------
        instruments :
            
        nb_bars :
            
        ts :
             (Default value = (4)
        4) :
            
        tempo :
             (Default value = 120)
        tpq :
             (Default value = 24)

        Returns
        -------

        """
        track_keys = [
            (idx, INSTRUMENTS_DICT[ins][0], INSTRUMENTS_DICT[ins][1] == 1)
            for idx, ins in enumerate(instruments)
        ]
        bars = [
            [
                0,
                0,
                "M",
                0,
                "",
                ts[0],
                ts[1],
            ]
            for _ in range(nb_bars)
        ]
        assert len(bars) == nb_bars, "Bars length mismatch"
        tracks = {}
        for track_key in track_keys:
            tracks[track_key] = [cls._get_fake_note_vector() for _ in range(nb_bars)]

            assert len(tracks[track_key]) == nb_bars, "Track and bar length mismatch"

        return cls(bars, tracks, track_keys, tempo, tpq=tpq)



import numpy as np
import time
import os
import requests
import tempfile
import warnings
from maidi.integrations.base import MidiApiIntegration
from maidi.integrations.api.musiclang import models
from maidi import MidiScore
from maidi import constants


# Enum of models

def get_musiclang_tag_list():
    from maidi.analysis.tags_providers import DensityTagsProvider, MinMaxPolyphonyTagsProvider,MinMaxRegisterTagsProvider, SpecialNotesTagsProvider
    tag_providers= [DensityTagsProvider, MinMaxPolyphonyTagsProvider, MinMaxRegisterTagsProvider, SpecialNotesTagsProvider]
    all_tags = []
    for tag_provider in tag_providers:
        provider = tag_provider()
        all_tags += provider.get_tags_list()

    return set(all_tags)


AVAILABLE_TAGS = get_musiclang_tag_list()


class MusicLangAPI(MidiApiIntegration):
    """
    Class to interact with the MusicLang API from a maidi.MidiScore

    **Usage:**


    >>> import os
    >>> import numpy as np
    >>> API_URL = os.getenv("API_URL")
    >>> API_KEY = os.getenv("API_KEY")
    >>> from maidi.integrations.api import MusicLangAPI
    >>> from maidi import MidiScore,instrument
    >>> api = MusicLangAPI(API_URL, API_KEY)
    >>> score = MidiScore.from_empty(instruments=[instrument.PIANO, instrument.ELECTRIC_BASS_FINGER], nb_bars=4, ts=(4, 4), tempo=120)
    >>> mask = np.ones((2, 4))
    >>> predicted_score = api.predict(score, mask, model="control_masking_large", timeout=120, temperature=0.95)
    >>> predicted_score.write("predicted_score.mid")


    """
    PREDICT_ASYNC_ENDPOINT = "predict_long"
    POLLING_ENDPOINT = "polling_predict"
    MAX_CONTEXT = 16 # Maximum context size of 16 bars for all masking models

    def __init__(self, api_url='https://api.musiclang.io', api_key=None, verbose=False):
        """
        Initialize the MusicLangAPI class

        Parameters
        ----------
        api_url : str
            URL of the API
        api_key : str
            Your secret API key to use
        verbose : bool
            If True, print debug information (Default value = False)

        """
        super().__init__()
        self.api_url = api_url
        self.api_key = api_key

        if self.api_key is None:
            # Get env
            self.api_key = os.getenv(constants.API_KEY_VARIABLE, None)
        if self.api_key is None:  # If not in env and not passed as a variable, raise an error
            raise ValueError('API key missing')

        self.verbose = verbose

    def _predict_with_api(
            self,
            score,
            mask,
            model="control_masking_large",
            temperature=0.95,
            async_mode=False,
            polling_interval=1,
            chords=None,
            tags=None,
            **prediction_kwargs,
    ):
        """

        Parameters
        ----------
        mask : np.array, shape (n_tracks, n_bars)
            The mask to use for the prediction
        model : str, optional, (Default value = "control_masking_large")
            The choice of model to use
        temperature : float, (Default value = 0.95)
            The temperature for the model, don't go higher than 1.0
        async_mode : bool, optional, (Default value = False)
            If True, return the task id, otherwise wait for the request to finish with polling
        polling_interval : int, optional, (Default value = 1)
            Interval in seconds to poll the API
        **prediction_kwargs :
            Additional arguments for the model (for example chord and control tags)

        Returns
        -------

        """
        mask = np.asarray(mask)
        score.check_mask(mask)
        task_id = self._call_predict_api(
            score.to_base64(), mask, model, temperature, chords, tags, **prediction_kwargs
        )
        if async_mode:
            return task_id
        else:
            score = self.from_task_id(task_id, polling_interval=polling_interval)
        return score

    def pprint(self, *args, **kwargs):
        if self.verbose:
            print(*args, **kwargs)


    def generate_from_scratch(self, instruments, nb_bars, ts, tempo, model=models.MODEL_CONTROL_MASKING_LARGE, chords=None, tags=None, temperature=0.95,
                              **kwargs):
        """Generate a score from scratch with the given parameters

        Parameters
        ----------
        instruments : list[str]
            List of instruments to use
        nb_bars : int
            Number of bars
        ts : tuple
            Time signature
        tempo : int
            Tempo
        model : str
            Model to use (Default value = models.MODEL_CONTROL_MASKING_LARGE)
        chords : list[tuple or None]
            List of chord tuples to guide the generation
        tags : list[list[list]] or None
            Tags to guide the generation
        temperature : float
            Temperature for the model (Default value = 0.95)
        **kwargs :
            Additional arguments for the model to pass to the `predict` method

        **Usage**

        >>> import numpy as np
        >>> from maidi import MidiScore, instrument
        >>> from maidi.integrations.api import MusicLangAPI
        >>> from maidi.integrations.api.musiclang import models
        >>> api = MusicLangAPI(API_URL, API_KEY)
        >>> score = api.generate_from_scratch(instruments=[instrument.ELECTRIC_PIANO, instrument.ALTO_SAX], nb_bars=4, ts=(4, 4), tempo=120, model=models.MODEL_CONTROL_MASKING_LARGE, temperature=0.95)
        >>> score.write("generated_score.mid")
        """
        score = MidiScore.from_empty(instruments=instruments, nb_bars=nb_bars, ts=ts, tempo=tempo)
        mask, new_tags, new_chords = score.get_empty_controls(prevent_silence=False)
        if tags is None:
            tags = new_tags
        if chords is None:
            chords = new_chords

        mask[:, :] = 1

        return self.predict(score, mask, model=model, temperature=temperature, chords=chords, tags=tags, **kwargs)

    def predict(
            self,
            score,
            mask,
            model=models.MODEL_CONTROL_MASKING_LARGE,
            timeout=120,
            temperature=0.95,
            cut_silenced_bars=False,
            regen_missing_bars=False,
            async_mode=False,
            polling_interval=1,
            tags=None,
            chords=None,
            **prediction_kwargs,
    ):
        """Predict the score with the given mask and prediction parameters

        **Usage**

        >>> import numpy as np
        >>> from maidi import MidiScore
        >>> from maidi.integrations.api import MusicLangAPI
        >>> score = MidiScore.from_empty(instruments=['piano'], nb_bars=4, ts=(4, 4), tempo=120)
        >>> mask = np.ones((1, 4))
        >>> api = MusicLangAPI(API_URL, API_KEY)  # Get your API_URL and API_KEY here
        >>> predicted_score = api.predict(score, mask, model="control_masking_large", timeout=120, temperature=0.95)
        >>> predicted_score.write("predicted_score.mid")

        Parameters
        ----------
        score : MidiScore
            The score to predict
        mask : np.array of shape (n_tracks, n_bars)
            The mask to use for the prediction
        model :
            str, model to use (for the API) (Default value = "control_masking_large")
        timeout :
            int, timeout for the API (Default value = 120)
        temperature :
            float, temperature for the model (Default value = 0.95)
        cut_silenced_bars :
            bool, if True cut silenced bars at the beginning and end (Default value = False)
        regen_missing_bars :
            bool, if True regenerate missing bars with another call to predict (Default value = False)
        prediction_kwargs :
            dict, additional arguments for the model (specifically `chord` and control `tags`)
        async_mode :
            bool, if True return the task id, otherwise wait for the request to finish with polling (Default value = False)
        polling_interval :
            int, interval in seconds to poll the API, only used if async_mode is False (Default value = 1)

        chords: list[tuple or None ]  or None
            List of tuple (chord_degree, tonality_degree, tonality_mode, roman numeral extension). Check the user guide for more information
        tags: list[list[list]] or None (n_tracks, n_bars, <variable length number of tags>)
            Way to specify soft constraints on the generation for each bar of each track, check the user guide for more information

        **prediction_kwargs :

        Returns
        -------
        type
            MidiScore or str : predicted score if sync mode

        """

        if model not in models.MODELS:
            raise ValueError(f"Model {model} not existing. Models available : {models.MODELS}")
        if temperature > 1.0:
            raise ValueError("Temperature must be lower than 1.0")
        if polling_interval < 0:
            raise ValueError("Polling interval must be > 0")
        if score.nb_bars > MusicLangAPI.MAX_CONTEXT:
            raise ValueError(f"The prompted score must be less than {MusicLangAPI.MAX_CONTEXT} bars.")

        if tags is not None:
            self._check_tags_exists(tags)

        if score.nb_bars != mask.shape[1]:
            raise ValueError(f'Wrong number of bars in the mask, (score) : {score.nb_bars} (mask): {mask.shape[1]}')

        if score.nb_tracks != mask.shape[0]:
            raise ValueError(f'Wrong number of tracks in the mask, (score) : {score.nb_tracks} (mask): {mask.shape[0]}')

        if chords is not None:
            self._check_chords_exists(chords)

        if chords is not None and (len(chords) != score.nb_bars):
            raise ValueError(f'Wrong number of chords (chords) :{len(chords)} (score) : {score.nb_bars}')

        score_to_predict = score.copy()
        mask = np.asarray(mask)
        score_to_predict.check_mask(mask)
        result = self._predict_with_api(
            score_to_predict,
            mask,
            model=model,
            timeout=timeout,
            temperature=temperature,
            async_mode=async_mode,
            polling_interval=polling_interval,
            chords=chords,
            tags=tags,
            **prediction_kwargs,
        )
        if async_mode:
            return result  # Return the task id

        score = result
        score = score.sanitize_score(score_to_predict, mask)
        # Cut silenced bars if at the end or beginning
        if cut_silenced_bars:
            score = score.cut_silenced_bars_at_beginning_and_end()

        # Regenerate missing bars if necessary
        if regen_missing_bars and score.nb_bars < mask.shape[1]:
            nb_missing_bars = mask.shape[1] - score.nb_bars
            score = score.add_silence_bars(
                mask.shape[1] - score.nb_bars, add_fake_notes=True
            )
            new_mask = score.get_mask()
            new_mask[:, -nb_missing_bars:] = 1
            return score.predict(
                new_mask,
                model=model,
                timeout=timeout,
                temperature=temperature,
                cut_silenced_bars=cut_silenced_bars,
                regen_missing_bars=regen_missing_bars,
            )
        return score


    def _get_subchords_and_subtags(self, chords, tags, index_start, index_end, offset=None):
        chords = chords[index_start:index_end] if chords is not None else None
        if tags is not None:
            tags = [[bar for bar in track[index_start:index_end]] for track in tags]

        if offset is not None:
            if chords is not None:
                chords = ([None] * offset) + chords
            if tags is not None:
                tags = [([[]] * offset) + track for track in tags]
        return chords, tags

    def create_transition(self, score1, score2, nb_bars_transition,
                          model=models.MODEL_CONTROL_MASKING_LARGE,
                          return_all=True,
                          chords=None,
                          tags=None,
                          timeout=120,
                          temperature=0.95,
                          async_mode=False,
                          polling_interval=1,
                          **prediction_kwargs):
        """
        Create a transition between two scores using the MusicLang API.

        Parameters
        ----------
        score1 : MidiScore
            The initial score from which to transition.
        score2 : MidiScore
            The final score to transition to.
        nb_bars_transition : int
            The number of bars to use for the transition.
        model : str, optional
            The model to use for the transition (default is "control_masking_large").
        return_all : bool, optional
            If True, return the full concatenated score with transition, otherwise only the transition part (default is True).
        chords : list[tuple or None], optional
            List of chord tuples to guide the transition.
        tags : list[list[list]], optional
            Tags to guide the transition, provided as a list of lists for each track and bar.
        timeout : int, optional
            Timeout for the API call (default is 120 seconds).
        temperature : float, optional
            Temperature parameter for the model (default is 0.95).
        async_mode : bool, optional
            If True, return the task id, otherwise wait for the request to finish with polling (default is False).
        polling_interval : int, optional
            Interval in seconds to poll the API for completion (default is 1 second).

        Returns
        -------
        MidiScore
            The score with the transition bars added, or only the transition part based on the return_all parameter.

        Raises
        ------
        ValueError
            If the model is not recognized or if the input parameters are invalid.
        """

        self._validate_transition_params(score1, score2, nb_bars_transition)

        nb_bars_score1, nb_bars_score2 = self._calculate_bars_to_take(score1, score2, nb_bars_transition)

        transition_segment = self._create_transition_segment(score1, score2, nb_bars_score1, nb_bars_transition,
                                                             nb_bars_score2)
        mask = self._create_transition_mask(score1.nb_tracks, transition_segment.nb_bars, nb_bars_score1,
                                            nb_bars_transition)
        transition_chords, transition_tags = self._adjust_chords_and_tags(chords, tags)

        transition_result = self.predict(
            transition_segment,
            mask,
            model=model,
            timeout=timeout,
            temperature=temperature,
            async_mode=async_mode,
            polling_interval=polling_interval,
            chords=transition_chords,
            tags=transition_tags,
            **prediction_kwargs
        )

        if async_mode:
            return transition_result  # Return the task id

        return self._concatenate_transition_results(score1, transition_result, score2, nb_bars_score1,
                                                    nb_bars_transition, nb_bars_score2, return_all)

    def _validate_transition_params(self, score1, score2, nb_bars_transition):
        if score1.nb_tracks != score2.nb_tracks:
            raise ValueError('The number of tracks in score1 and score2 must be the same')
        if score1.instruments != score2.instruments:
            warnings.warn(
                'Instruments are different in the two scores, interpreting score2 instruments as score1 instruments')

        if nb_bars_transition > self.MAX_CONTEXT - 4:
            raise ValueError(f"Transition size must be less than {self.MAX_CONTEXT - 4} bars")

    def _calculate_bars_to_take(self, score1, score2, nb_bars_transition):
        nb_bars_score1 = min((self.MAX_CONTEXT - nb_bars_transition) // 2, score1.nb_bars)
        nb_bars_score2 = min(self.MAX_CONTEXT - nb_bars_transition - nb_bars_score1, score2.nb_bars)
        return nb_bars_score1, nb_bars_score2

    def _create_transition_segment(self, score1, score2, nb_bars_score1, nb_bars_transition, nb_bars_score2):
        return score1[:, -nb_bars_score1:].add_silence_bars(nb_bars_transition).concatenate(score2[:, :nb_bars_score2],axis=1)

    def _create_transition_mask(self, nb_tracks, nb_bars, nb_bars_score1, nb_bars_transition):
        mask = np.zeros((nb_tracks, nb_bars))
        mask[:, nb_bars_score1:nb_bars_score1 + nb_bars_transition] = 1
        return mask

    def _adjust_chords_and_tags(self, chords, tags):
        return chords, tags

    def _concatenate_transition_results(self, score1, transition_result, score2, nb_bars_score1, nb_bars_transition,
                                        nb_bars_score2, return_all):
        if return_all:
            final_score = score1[:, :-nb_bars_score1].concatenate(transition_result, axis=1).concatenate(
                score2[:, nb_bars_score2:], axis=1)
        else:
            final_score = transition_result[:, nb_bars_score1:nb_bars_score1 + nb_bars_transition]
        return final_score

    def extend(self, score, nb_bars_added,
               model=models.MODEL_CONTROL_MASKING_LARGE,
               nb_added_bars_step=None,
               chords=None,
               tags=None,
               timeout=120,
               temperature=0.95,
               polling_interval=1,
               **prediction_kwargs):
        """
        Extend the given score by a specified number of bars using the MusicLang API.

        Parameters
        ----------
        score : MidiScore
            The initial score to extend.
        nb_bars_added : int
            The number of bars to add to the score.
        model : str, optional
            The model to use for the extension (default is "control_masking_large").
        nb_added_bars_step : int, optional
            Number of bars to add in each step (if None, the extension will use the maximum context size).
        chords : list[tuple or None], optional
            List of chord tuples to guide the extension.
        tags : list[list[list]], optional
            Tags to guide the extension, provided as a list of lists for each track and bar.
        timeout : int, optional
            Timeout for the API call (default is 120 seconds).
        temperature : float, optional
            Temperature parameter for the model (default is 0.95).
        polling_interval : int, optional
            Interval in seconds to poll the API for completion (default is 1 second).
        **prediction_kwargs : dict
            Additional arguments for the model

        Returns
        -------
        MidiScore
            The extended score with the specified number of additional bars.

        Raises
        ------
        ValueError
            If the model is not recognized or if the input parameters are invalid.
        """

        new_score = score.copy()
        nb_bars_current = new_score.nb_bars
        offset_bar = 0

        while new_score.nb_bars < (nb_bars_current + nb_bars_added):
            remaining_bars = (nb_bars_current + nb_bars_added) - new_score.nb_bars
            nb_added_bars_in_this_step = self._determine_bars_step(nb_added_bars_step, remaining_bars)
            new_score = new_score.add_silence_bars(nb_added_bars_in_this_step, add_fake_notes=True)
            predicted_score = new_score[:, -self.MAX_CONTEXT:]
            base_score = new_score[:, :-predicted_score.nb_bars]

            mask = self._create_mask(predicted_score, nb_added_bars_in_this_step)
            subchords, subtags = self._get_subchords_and_subtags(chords, tags, offset_bar,
                                                                 offset_bar+nb_added_bars_in_this_step,
                                                                 mask.shape[1] - nb_added_bars_in_this_step)

            offset_bar += nb_added_bars_in_this_step
            result_predicted_score = self._predict_segment(predicted_score, mask, model, subtags, subchords,
                                                           temperature, polling_interval, timeout, prediction_kwargs)
            new_score = base_score.concatenate(result_predicted_score, axis=1)

        return new_score

    def _determine_bars_step(self, nb_added_bars_step, remaining_bars):
        if nb_added_bars_step is not None:
            return min(remaining_bars, nb_added_bars_step)
        else:
            return min(self.MAX_CONTEXT, remaining_bars)

    def _create_mask(self, predicted_score, nb_added_bars_in_this_step):
        mask = predicted_score.get_mask()
        mask[:, -nb_added_bars_in_this_step:] = 1
        return mask

    def _predict_segment(self, predicted_score, mask, model, subtags, subchords,
                         temperature, polling_interval, timeout, prediction_kwargs):
        return self.predict(predicted_score,
                            mask,
                            tags=subtags,
                            model=model,
                            chords=subchords,
                            temperature=temperature,
                            polling_interval=polling_interval,
                            timeout=timeout,
                            prediction_kwargs=prediction_kwargs)




    def _check_tags_exists(self, tags):
        all_tags = set([el for track in tags for bar in track for el in bar])

        if not AVAILABLE_TAGS.issuperset(all_tags):
            tags_not_existing = all_tags - AVAILABLE_TAGS
            raise ValueError(f'Unexisting tags have been used : {tags_not_existing} This is the list of available tags {AVAILABLE_TAGS}')

    def _check_chords_exists(self, chords):
        from maidi.chords_symbols import is_valid_extension, ALL_EXTENSIONS
        def check_chord_exist(chord, index_chord):
            deg, tone, mode, extension = chord
            if not isinstance(deg, int) or deg < 0 or deg > 6:
                raise ValueError(f'Chord degree is wrong : {deg}. It should be an integer between 0 and 6 included')
            if not isinstance(tone, int) or tone < 0 or tone > 11:
                raise ValueError(f'Chord degree is wrong : {tone}. It should be an integer between 0 and 11 included')
            if not (mode in ['m', 'M']):
                raise ValueError(f'Scale mode is wrong, it should be either "m" (minor mode) or "M" (major mode)')
            if not is_valid_extension(extension):
                raise ValueError(f'Wrong extension : {extension}, must be in {ALL_EXTENSIONS} with optionally added "(sus2)" or "(sus4)"')

        if chords is None:
            return
        for idx, chord in enumerate(chords):
            if chord is not None:
                check_chord_exist(chord, index_chord=idx)



    def _call_polling_api(self, task_id):
        """

        Parameters
        ----------
        url :

        key :

        task_id :


        Returns
        -------

        """
        url = os.path.join(self.api_url, self.POLLING_ENDPOINT, task_id)
        headers = {"x-api-key": self.api_key}
        response = requests.get(url, headers=headers)
        response.raise_for_status()
        return response.json()

    def from_task_id(self, task_id, polling_interval=1, timeout=3000):
        """Create a score from a task id by calling the API polling endpoint each polling_interval seconds
        until completion or failure.

        **Usage**

        >>> import numpy as np
        >>> import time
        >>> from maidi import MidiScore
        >>> from maidi.integrations.api import MusicLangAPI
        >>> score = MidiScore.from_empty(instruments=['piano'], nb_bars=4, ts=(4, 4), tempo=120)
        >>> mask = np.ones((1, 4))
        >>> api = MusicLangAPI(API_URL, API_KEY)
        >>> task_id = api.predict(score, mask, async_mode=True)
        >>> score = api.from_task_id(task_id, polling_interval=3) # Wait for the task to complete by polling the API every 3 seconds

        Parameters
        ----------
        task_id :
            str, task identifier you got from the predict call
        polling_interval :
            int, interval in seconds to poll the API (Default value = 1)
        timeout :
            int, timeout in seconds, raise an error if the task is not completed after this time (Default value = 3000)

        Returns
        -------

        """
        total_time = 0
        # Poll the API
        while True:
            if total_time > timeout:
                raise TimeoutError("Timeout reached")
            result = self._call_polling_api(task_id)
            if result["status"] == "COMPLETED":
                break
            elif result["status"] == "FAILED":
                raise ValueError("Task failed")

            self.pprint(f"Task {task_id} not completed yet, waiting {polling_interval} seconds")
            time.sleep(polling_interval)
            total_time += polling_interval

        # Download the file
        score = MidiScore.from_base64(result["result"]["midi"])

        return score


    def poll_api(self, task_id):
        """Poll the API with a task id once

        Parameters
        ----------
        task_id :
            str, task identifier you got from the predict call

        Returns
        -------

        result: None or MidiScore

        Raises
        ------
        ValueError
            Task failed

        """
        # Poll the API
        result = self._call_polling_api(task_id)
        if result["status"] == "COMPLETED":
            self.pprint(f"Task {task_id} successfully completed")
            return MidiScore.from_base64(result["result"]["midi"])
        elif result["status"] == "FAILED":
            raise ValueError("Task failed")
        self.pprint(f"Task {task_id} not completed yet")

        return None

    def _call_predict_api(
            self, midi_base64, mask, model, temperature, chords, tags, **prediction_kwargs
    ):
        """
        Protected method that actually prepare the payload and call the API

        Parameters
        ----------
        midi_base64 : str
            Base64-encoded midi file

        mask : np.Array
            Array representing which bar of which track to regenerate

        model : str
         Model in model list, check user guide.

        temperature : float
            Temperature of the model, between 0 and 1.0

        **prediction_kwargs : dict
            Other model arguments

        Returns
        -------
            task_id : str

        """
        url = os.path.join(self.api_url, self.PREDICT_ASYNC_ENDPOINT)
        payload = {
            "model": model,
            "temperature": temperature,
            "mask": mask.tolist(),
            "file": midi_base64,
        }
        if tags is not None:
            payload['tags'] = tags
        if chords is not None:
            payload['chords'] = chords

        payload.update(prediction_kwargs)
        headers = {"x-api-key": self.api_key}
        response = requests.post(url, json=payload, headers=headers)
        response.raise_for_status()
        return response.json()["task_id"]



    def _predict_with_predictor(
            self,
            score,
            predictor,
            mask,
            model="large",
            timeout=3000,
            temperature=0.95,
            **prediction_kwargs,
    ):
        """

        Not implemented yet.

        Parameters
        ----------
        predictor :

        mask :

        model :
             (Default value = "large")
        timeout :
             (Default value = 3000)
        temperature :
             (Default value = 0.95)
        **prediction_kwargs :


        Returns
        -------

        """
        mask = np.asarray(mask)
        score.check_mask(mask)
        prompt_file, output_file, midi_file, output_midi_file = (
            self._create_temp_midi_files(score)
        )
        score.write(midi_file)
        predictor(
            midi_file,
            output_midi_file,
            prompt_file,
            output_file,
            mask,
            chord_range=(0, mask.shape[1]),
            model=model,
            timeout=timeout,
            temperature=temperature,
            **prediction_kwargs,
        )
        score = MidiScore.from_midi(output_midi_file)
        # Clean up the files
        self._remove_temp_midi_files(
            prompt_file, output_file, midi_file, output_midi_file
        )
        return score

    def _create_temp_midi_files(self, score):
        """ """
        prompt_file = tempfile.NamedTemporaryFile(suffix=".txt").name
        output_file = tempfile.NamedTemporaryFile(suffix=".txt").name
        midi_file = tempfile.NamedTemporaryFile(suffix=".mid").name
        output_midi_file = tempfile.NamedTemporaryFile(suffix=".mid").name
        score.write(midi_file)
        return prompt_file, output_file, midi_file, output_midi_file

    def _remove_temp_midi_files(
            self, prompt_file, output_file, midi_file, output_midi_file
    ):
        """

        Parameters
        ----------
        prompt_file :

        output_file :

        midi_file :

        output_midi_file :


        Returns
        -------

        """
        os.remove(prompt_file)
        os.remove(output_file)
        os.remove(midi_file)
        os.remove(output_midi_file)


from maidi.analysis import TagsProvider

REGISTERS = {
    "highest": (81, 130),
    "sopranissimo": (74, 95),
    "soprano": (60, 81),
    "mezzo_soprano": (57, 78),
    "alto": (55, 76),
    "contralto": (53, 74),
    "tenor": (48, 69),
    "baritone": (45, 65),
    "bass": (41, 64),
    "contrabass": (36, 57),
    "lowest": (0, 36),
}

ORDERED_FROM_LOWEST = list(
    sorted(REGISTERS.keys(), key=lambda x: REGISTERS[x][0], reverse=True)
)
ORDERED_FROM_HIGHEST = list(sorted(REGISTERS.keys(), key=lambda x: REGISTERS[x][1]))

LOW_PITCHES = [REGISTERS[x][0] for x in ORDERED_FROM_LOWEST]






def get_tags_names():
    """

    Returns
    -------

    """
    min_tokens = [
        "CONTROL_MIN_REGISTER__" + register for register in ORDERED_FROM_LOWEST
    ]
    max_tokens = [
        "CONTROL_MAX_REGISTER__" + register for register in ORDERED_FROM_HIGHEST
    ]
    return min_tokens + max_tokens


class MinMaxRegisterTagsProvider(TagsProvider):
    ALL_TAGS = get_tags_names()

    def get_tags(self, track_bar, chord, score):
        pitches = self.get_pitches(track_bar, chord, score)
        return get_min_max_register_tags(pitches)



def get_min_max_registers(pitches: list[int]) -> list[str]:
    """Get highest register that encompasses min_pitch, and lowest register that encompasses max_pitch

    Parameters
    ----------
    pitches :
        List of pitch values
    pitches: list[int] :
        

    Returns
    -------
    type
        Tuple containing the highest and lowest registers

    """
    pitches = set(pitches)
    if len(pitches) == 0:
        return []

    min_pitch = min(pitches)
    max_pitch = max(pitches)

    min_register = "tenor"  # Assuming this is the lowest register
    max_register = "tenor"  # Assuming this is the highest register

    for register in ORDERED_FROM_LOWEST:
        if REGISTERS[register][0] <= min_pitch:
            min_register = register
            break

    if max_pitch <= REGISTERS[min_register][1]:
        return [min_register, min_register]

    for register in ORDERED_FROM_HIGHEST:
        if REGISTERS[register][1] >= max_pitch:
            max_register = register
            break

    return [min_register, max_register]


def get_min_max_register_tags(pitches: list[int]) -> list[str]:
    """

    Parameters
    ----------
    pitches: list[int] :


    Returns
    -------

    """
    registers = get_min_max_registers(pitches)
    if len(registers) == 0:
        return []
    return [
        "CONTROL_MIN_REGISTER__" + registers[0],
        "CONTROL_MAX_REGISTER__" + registers[1],
    ]


def get_token_names():
    """ """
    min_tokens = [
        "CONTROL_MIN_REGISTER__" + register for register in ORDERED_FROM_LOWEST
    ]
    max_tokens = [
        "CONTROL_MAX_REGISTER__" + register for register in ORDERED_FROM_HIGHEST
    ]
    return min_tokens + max_tokens
